"""Map"""

from __future__ import annotations

import time as keep_time
from functools import cached_property
from operator import is_
from typing import TYPE_CHECKING

from gana import sigma

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._x import _X
    from ..indices.domain import Domain
    from ..variables.aspect import Aspect


class Map:
    """Maps between domains

    :param aspect: Aspect to which the constraint is applied.
    :type aspect: Aspect
    :param domain: Domain over which the aspect is defined.
    :type domain: Domain
    :param reporting: If True, the map is for a reporting variable.
    :type reporting: bool
    """

    def __init__(
        self, aspect: Aspect, domain: Domain, label: str = "", reporting=False
    ):

        # if the variable is being defined for the first time, do not bother with the rest
        # Also note that if a variable already exists then a new is not created
        # thus map_domain is not called in Bind.V()
        # Essentially, the domain being added is always new!

        # 1. a new domain is being added to a new disposition
        # e.g.:
        #   (water, dam, goa, year) with disposition ['resource', 'operation', 'space', 'time'] exists
        #   (dam, goa, q) with disposition ['operation', 'space', 'time'] is added

        # # 2. a new domain is being added to an existing disposition
        # # e.g.:
        # #   (dam, goa, year) with disposition ['operation', 'space', 'time'] exists
        # #   (dam, goa, q) with disposition ['operation', 'space', 'time'] is added
        # #   The variable will need to be mapped across 'time

        # mapping happens between:
        # 1. sparser time to denser time in same space
        # 2. contained location to parent location
        # 3. (sparser time, contained location) to (denser time, parent location)
        # you cant map at the same level
        # consider the State Goa, with two towns Madgaon and Ponje
        # Goa with Maharashtra, make up India (for the purpose of modeling)
        # Aspect from Goa can be mapped to India
        # Goa does not need to be mapped to Maharashtra
        # Madgaon and Ponje are both mapped to Just Goa
        # note that .alsohas() checks at lower levels
        # the operators <, > wont work since they will add any space a higher order
        # this will lead to adding twice. India = Goa + Madgaon + Ponje (WRONG)
        # We scale only one level up

        self.aspect = aspect
        self.domain = domain
        self.label = label
        self.reporting = reporting

        if self.domain.lag:
            return

        self.model = self.aspect.model
        self.dispositions = self.model.dispositions
        self.program = self.model.program

        # this is the disposition of the variable to be mapped
        # through time and space
        time, space = self.domain.periods, self.domain.space

        # these are periods denser and sparser than the current domain
        denser_periods, sparser_periods = self.model.time.split(time)

        # these are spaces contained in location and parent location to which this location belongs
        # this gives all the dispositions at which the aspect has been defined
        dispositions = self.dispositions[self.aspect][self.domain.primary]

        if space not in dispositions or time not in dispositions[space]:
            return

        # --- Time mapping ---
        self._map_across_time(
            dispositions, space, time, sparser_periods, denser_periods
        )

        # --- Space mapping ---
        contained_locs, parent_loc = self.model.space.split(space)
        self._map_across_space(dispositions, contained_locs, parent_loc, time)

        # --- Bind mapping ---
        self._map_across_binds(dispositions, space, time)

        # --- Mode mapping ---
        self._map_across_modes()

    @cached_property
    def name(self) -> str:
        return f"{self.aspect.name}_map"

    @cached_property
    def maps(self) -> dict[str, list[Domain]]:
        return self.aspect.maps_report if self.reporting else self.aspect.maps

    # ---------------------------------------------------------------------- #
    # Helper functions
    # ---------------------------------------------------------------------- #

    def _map_across_time(
        self, dispositions, space, time, sparser_periods, denser_periods
    ):
        for sp in sparser_periods:
            # check if the aspect has been defined for a sparser period
            # this creates a map from this domain to a sparser domain
            if sp in dispositions[space] and is_(sp.of, time):
                self.writecons_map(
                    self.domain, self.domain.change({"periods": sp}), tsum=True
                )

        for dp in denser_periods:
            if dp in dispositions[space] and is_(time.of, dp):
                binds_dict = dispositions[space][dp]
                # TODO - check this
                # here I am re creating Bind objects
                # from the dict of the form {aspect: {component: {aspect: {component: {...}}}}}
                # there has to be a way to avoid this
                # I make a list  of bounds as such [aspect(component), aspect(component), ...]
                binds = [
                    aspect(component)
                    for aspect, comp_dict in binds_dict.items()
                    for component in comp_dict
                ]
                from_domain = self.domain.copy()
                from_domain.periods, from_domain.binds = dp, binds
                self.writecons_map(from_domain, self.domain, tsum=True)

    def _map_across_space(self, dispositions, contained_locs, parent_loc, time):

        parent_loc = self.domain.location.isin

        if parent_loc:
            if parent_loc not in dispositions or time not in dispositions[parent_loc]:
                return
            self.writecons_map(
                self.domain,
                self.domain.change({"location": parent_loc}),
            )

        # for location in contained_locs:
        #     if location not in dispositions or time not in dispositions[location]:
        #         # the aspect is already defined at this location
        #         # we need to check if it is defined for any periods sparser than time
        #         continue
        #     # binds = dispositions[location][time]
        #     # if not binds:

        #     print('asdadada', location, self.domain)
        #     self.writecons_map(self.domain.change({"location": location}), self.domain)
        # else:
        #     new_binds = [k(list(v)[0]) for k, v in binds.items()]
        #     print('aaaa', self.aspect, new_binds, self.domain)
        #     # consider the case where overall consumption for water in some location and time is defined
        #     # now user defines consumption due to using cement during construction
        #     # we should have the constraint consume(water, goa, 2025) = consume(water, goa, 2025, use, cement)
        #     # in that case, binds is just []
        #     self.writecons_map(
        #         self.domain.change({"location": location, "binds": new_binds}),
        #         self.domain,
        #     )

        # if parent_loc in dispositions:

        #     self.writecons_map(
        #         self.domain, self.domain.change({"location": parent_loc})
        #     )

    def _map_across_binds(self, dispositions, space, time):
        if self.domain.binds or not dispositions[space][time]:
            return
        # if the current variable being declared has no binds
        # but the aspect has already been defined at this location and time with binds
        # there is a need to map from the defined binds to no binds
        # get list of domains of the aspect

        domains = [
            d
            for d in self.aspect.domains
            if is_(d.primary, self.domain.primary)
            and is_(d.space, space)
            and is_(d.periods, time)
            and d.binds
        ]
        for domain in domains:
            self.writecons_map(domain, self.domain)

    def _map_across_modes(self):
        # modes need some additional checks
        # constraints could be written using parent modes or child modes
        # these checks avoid writing the same constraint twice
        # AVOIDS: v_t = v_t,m1 + v_t,m2 + v_t,m; v_t,m = v_t,m1 + v_t,m2
        # CORRECT: v_t = v_t,m; v_t,m = v_t,m1 + v_t,m2 OR v_t = v_t,m1 + v_t,m2
        if not self.domain.modes:
            return
        if self.domain.modes.parent:
            self.writecons_map(self.domain, self.domain.change({"modes": None}))
        else:
            self.writecons_map(
                self.domain, self.domain.change({"modes": None}), msum=True
            )

    def _give_cname(
        self,
        var,
        from_domain,
        to_domain,
        tsum: bool = False,
        msum: bool = False,
    ) -> str:
        """Return canonical map constraint name based on domain relationship and aggregation type."""
        if tsum:
            return f"{var}{from_domain.idxname}_to_{to_domain.idxname}_tmap"

        if msum:
            # Original behavior: use from_domain idxname for per-mode naming
            return f"{var}{from_domain.idxname}_mmap"

        if from_domain.modes:
            if from_domain.modes.parent:
                parent_domain = from_domain.change({"modes": from_domain.modes.parent})
                return f"{var}{parent_domain.idxname}_mmap"
            return f"{var}{from_domain.idxname}_mmap"

        return f"{var}{to_domain.idxname}_map"

    def _give_sum(self, domain: Domain, tsum=False, msum=False):
        """Gives the sum of the variable over the domain"""
        varname = (
            f"x_{self.aspect.name}" if (msum and self.reporting) else self.aspect.name
        )
        v = getattr(self.program, varname)

        if tsum:
            # if the domain has been mapped to but this is a time sum
            # we need to first map time
            # and then add it to an existing map at a lower domain
            return sigma(v(*domain.Ilist), domain.time.I)
        if msum:
            # if the domain has been mapped to but this is a mode sum
            # we need to first map modes
            # and then add it to an existing map at a lower domain
            return sigma(v(*domain.Ilist), domain.modes.I)
        # the copy is important since otherwise, the printing will take
        # the update index if the variable is mutated

        return v(*domain.Ilist).copy()

    # ---------------------------------------------------------------------- #
    # Constraint writing
    # ---------------------------------------------------------------------- #
    def writecons_map(
        self, from_domain: Domain, to_domain: Domain, tsum=False, msum=False
    ):
        """Scales up variable to a lower dimension"""

        if to_domain not in self.maps:
            self.maps[to_domain] = [from_domain]
            exists = False
        elif from_domain in self.maps[to_domain]:
            return
        else:
            self.maps[to_domain].append(from_domain)
            exists = True

        var = self.aspect.reporting if self.reporting else self.aspect

        rhs = self._give_sum(from_domain, tsum, msum)

        cname = self._give_cname(var, from_domain, to_domain, tsum, msum)

        print(f"--- Mapping {var}: {from_domain} → {to_domain}")
        start = keep_time.time()

        v_lower = self(*to_domain).X() if self.reporting else self(*to_domain).V()

        if not tsum and not msum and exists:
            cons_existing: C = getattr(self.program, cname)
            setattr(self.program, cname, cons_existing - rhs)

        else:
            cons = v_lower == rhs
            setattr(self.program, cname, cons)
            cons.categorize("Mapping")

        end = keep_time.time()
        print(f"    Completed in {end-start:.3f}s")
        if cname not in self.aspect.constraints:
            self.aspect.constraints.append(cname)
        from_domain.update_cons(cname)

    def __call__(self, *index: _X):
        return self.aspect(*index)



"""Fetch data from NREL's NSRDB database"""

from numpy import array, average
from pandas import DataFrame, to_datetime
from scipy.spatial import cKDTree

try:
    import h5pyd

    import_all = False
except ImportError:
    import_all = True


def fetch_nsrdb_data(
    attrs: list[str],
    year: int,
    lat_lon: tuple[float] | None = None,
    state: str = "",
    county: str = "",
    resolution: str = "",
    get: str = "max-population",
    save: str = None,
) -> DataFrame | tuple:
    """fetches nsrdb data from nearest coordinates (latitude, longitude)
    or from county in a state matching a particular 'get' metric

    Args:
        attrs (list[str]): attributes to fetch ['air_temperature', 'clearsky_dhi', 'clearsky_dni', 'clearsky_ghi', 'cloud_type', 'coordinates', 'dew_point', 'dhi', 'dni', 'fill_flag', 'ghi', 'meta', 'relative_humidity', 'solar_zenith_angle', 'surface_albedo', 'surface_pressure', 'time_index', 'total_precipitable_water', 'wind_direction', 'wind_speed']
        year (int): year of choice, e.g. 2019
        lat_lon (Tuple[float], optional): (latitude, longitude) to fetch closest data point. Defaults to None.
        state (str, optional): capitalized state name, e.g. 'Texas' . Defaults to ''.
        county (str, optional): capitalized county name, e.g. 'Brazos' . Defaults to ''.
        resolution (str, optional): choose from 'halfhourly', 'hourly', 'daily'. Defaults to ''.
        get (str, optional): Defaults to 'max-population'. From within county choose the data point that matches one of the following. 'max-population', 'max-elevation', 'max-landcover' 'min-population', 'min-elevation', 'min-landcover'

    Returns:
        DataFrame, tuple: Dataframe with output data, (latitude, longitude)
    """

    if import_all:
        print(
            "This is an optional feature. Please install h5pyd, or pip install energiapy[all]",
        )
        return None

    # fetches nsrdb data for the year
    nsrdb_data = h5pyd.File(f"/nrel/nsrdb/v3/nsrdb_{year!s}.h5", "r")
    time_index = to_datetime(nsrdb_data["time_index"][...].astype(str))

    if lat_lon is not None:
        # get coordinates for all locations
        coords = nsrdb_data["coordinates"][...]

        tree = cKDTree(coords)

        # find the data point closest to latitude and longitude
        def nearest_site(tree, latitude, longitude):
            lat_lon_query = array([latitude, longitude])
            # dist, pos = tree.query(lat_lon_query)
            return tree.query(lat_lon_query)[1]

        idx = nearest_site(tree=tree, latitude=lat_lon[0], longitude=lat_lon[1])

    else:
        # gets coordinates and associated data
        meta = DataFrame(nsrdb_data["meta"][...])
        # data matching state coordinates
        state_data = meta.loc[meta["state"] == str.encode(state)]
        county_data = state_data.loc[
            state_data["county"] == str.encode(county)
        ]  # data matching county

        # splits the get string, e.g. max - population, gives [max,
        # population(get_metric)]
        get_metric = get.split("-")[1]

        if get.split("-")[0] == "min":
            latitude = float(
                county_data["latitude"][
                    county_data[get_metric] == min(county_data[get_metric])
                ].iloc[0],
            )
            longitude = float(
                county_data["longitude"][
                    county_data[get_metric] == min(county_data[get_metric])
                ].iloc[0],
            )
            loc_data = county_data.loc[
                (county_data["latitude"] == latitude)
                & (county_data["longitude"] == longitude)
            ]

        if get.split("-")[0] == "max":
            latitude = float(
                county_data["latitude"][
                    county_data[get_metric] == max(county_data[get_metric])
                ].iloc[0],
            )
            longitude = float(
                county_data["longitude"][
                    county_data[get_metric] == max(county_data[get_metric])
                ].iloc[0],
            )
            loc_data = county_data.loc[
                (county_data["latitude"] == latitude)
                & (county_data["longitude"] == longitude)
            ]

        idx = loc_data.index[0]
        lat_lon = (latitude, longitude)

    timestep_dict = {
        "halfhourly": 1,  # native data set at 30 mins
        "hourly": 2,  # averages over the hour
        "daily": 48,  # averages over the day
    }
    averaged_output = DataFrame()

    psm_scale_dict = {
        attr: nsrdb_data[attr].attrs["psm_scale_factor"] for attr in attrs
    }

    for attr in attrs:
        full_output = nsrdb_data[attr][:, idx]  # native data set at 30 mins
        averaged_output[attr] = average(
            full_output.reshape(-1, timestep_dict[resolution]),
            axis=1,
        )  # averages over resolution
    averaged_output = averaged_output.set_index(
        time_index[:: timestep_dict[resolution]],
    )

    for attr in attrs:
        averaged_output[attr] = averaged_output[attr] / psm_scale_dict[attr]

    if save is not None:
        averaged_output.to_csv(save + ".csv")

    return lat_lon, averaged_output



"""Control Variable"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Self

from ..constraints.vmap import Map
from .aspect import Aspect

if TYPE_CHECKING:
    from ..indices.domain import Domain
    from .states import State


@dataclass
class Control(Aspect):
    """Some action to be performed by a component
    that adjusts a state variable
    can have a consequence (Impact)
    or elicit a motion (Stream) that has a consequence (Impact)
    """

    def __post_init__(self):
        Aspect.__post_init__(self)

        # a state that is being controlled
        self.state: State = None
        # a stream that is being controlled
        self.stream: Self = None
        # an impact that is being controlled
        self.impact: Self = None
        # what increase the state/impact/stream
        self.more: Self = None
        # what decreases the state/impact/stream
        self.less: Self = None

    def update(self, domain: Domain):
        """Add a domain to the decision variable"""

        # Write a mapping constraint
        Map(aspect=self, domain=domain)



"""Currency"""

from __future__ import annotations

from dataclasses import dataclass
from operator import is_
from typing import TYPE_CHECKING, Self

from ._commodity import _Commodity

if TYPE_CHECKING:
    from ...components.spatial.location import Location


@dataclass
class Currency(_Commodity):
    """Same as Economic Impact (Eco)

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """

    locs: list[Location] = None

    def __post_init__(self):
        _Commodity.__post_init__(self)

        # dictionary of exchange rates
        self.exchange = {}

        if self.locs:
            for loc in self.locs:
                loc.currency = self
                for l in loc.alsohas:
                    l.currency = self

    def howmany(self, cash: Self):
        """Exchange rate basically"""

        if is_(cash, self):
            return 1
        if cash in self.exchange:
            return self.exchange[cash]

        # # find a common currency
        # if list(self.conversion)[0] == list(cash.conversion)[0]:
        #     return (
        #         self.conversion[list(self.conversion)[0]]
        #         / cash.conversion[list(cash.conversion)[0]]
        #     )
        raise ValueError(f"{cash} does not have an exchange rate set {self.name}")

    def __eq__(self, other):
        if isinstance(other, Currency):
            self.exchange[other] = 1.0
        # assume it is a Conversion
        else:
            currency = list(other.conversion.keys())[0]
            rate = other.conversion[currency]

            # set the exchange rate of self against other
            self.exchange[currency] = rate

            for ex in currency.exchange:
                if ex not in self.exchange:
                    self.exchange[ex] = rate / currency.exchange[ex]

                    if self not in ex.exchange:

                        ex.exchange[self] = ex.exchange[currency] / rate

                    # if self not in ex.exchange:
                    #     ex.exchange[self] = currency.exchange[ex] / rate

            # set the exchange rate of other against self
            currency.exchange[self] = 1 / rate



"""System"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .._core._dimension import _Dimension

if TYPE_CHECKING:
    from ..components.commodity.currency import Currency
    from ..components.commodity.emission import Emission
    from ..components.commodity.land import Land
    from ..components.commodity.material import Material
    from ..components.commodity.resource import Resource
    from ..components.game.couple import Couple
    from ..components.game.player import Player
    from ..components.operation.process import Process
    from ..components.operation.storage import Storage
    from ..components.operation.transport import Transport


@dataclass
class System(_Dimension):
    """System representation as a Resource Task Network (RTN)
    All resources and tasks are attached to this object

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, generated based on the class and model name.
    :vartype name: str
    :ivar resources: List of resources. Defaults to [].
    :vartype resources: list[Resource]
    :ivar lands: List of land resources. Defaults to [].
    :vartype lands: list[Land]
    :ivar materials: List of material resources. Defaults to [].
    :vartype materials: list[Material]
    :ivar currencies: List of currency resources. Defaults to [].
    :vartype currencies: list[Currency]
    :ivar emissions: List of emission resources. Defaults to [].
    :vartype emissions: list[Emission]
    :ivar processes: List of processes. Defaults to [].
    :vartype processes: list[Process]
    :ivar storages: List of storage operations. Defaults to [].
    :vartype storages: list[Storage]
    :ivar transports: List of transport operations. Defaults to [].
    :vartype transports: list[Transport]
    :ivar players: List of players. Defaults to [].
    :vartype players: list[Player]
    :ivar couples: List of couples. Defaults to [].
    :vartype couples: list[Couple]
    """

    def __post_init__(self):
        # ---------- Commodities/Resources ----------

        self.resources: list[Resource] = []
        self.lands: list[Land] = []
        self.materials: list[Material] = []
        self.currencies: list[Currency] = []
        self.emissions: list[Emission] = []

        # ---------- Operations/Tasks ----------

        self.processes: list[Process] = []
        self.storages: list[Storage] = []
        self.transports: list[Transport] = []

        # ------------Decision-Makers----------------------
        self.players: list[Player] = []
        self.couples: list[Couple] = []

    @property
    def operations(self) -> list[Process | Storage | Transport]:
        """All Operations"""
        return self.processes + self.storages + self.transports



"""Calculation"""

from __future__ import annotations

from typing import TYPE_CHECKING, Self

from ...utils.math import normalize
from ..parameters.value import Value

if TYPE_CHECKING:
    from gana.block.program import Prg
    from gana.sets.constraint import C

    from ..._core._x import _X
    from ...represent.model import Model
    from ..variables.sample import Sample


class Calculate:
    """
    Calculate the value of a dependent variable based on the value of another variable.

    Usually used when another variable has a degree of freedom.

    :param calc: Calculated (derivative) variable bound.
    :type calc: Bind
    :param decision: Decision variable bound.
    :type decision: Bind

    :ivar decision: Decision variable bound.
    :vartype decision: Bind
    :ivar calc: Calculated (derivative) variable bound.
    :vartype calc: Bind
    :ivar domain: Domain of the calculated variable.
    :vartype domain: Domain
    :ivar name: Name of the calculated variable.
    :vartype name: str
    :ivar program: Program object to which the variable belongs.
    :vartype program: Prg
    :ivar index: Index set of the calculated variable.
    :vartype index: I
    """

    def __init__(
        self,
        calculation: Sample,
        sample: Sample,
    ):
        self.calculation = calculation

        self.model = self.calculation.model
        self.domain = self.calculation.domain
        self.name = self.calculation.name
        self.program = self.calculation.program
        self.index = self.calculation.index

        self.sample = sample

        self._forall: list[_X] = []

        # if nominal is provided
        # and multiplied by the nominal value
        self._nominal: float = None
        # the input argument is normalized if True
        self._normalize: bool = False

    def forall(self, index) -> Self:
        """Write the constraint for all indices in the index set"""
        self._forall = index
        return self

    def prep(self, nominal: float = 1, norm: bool = True) -> Self:
        """Nominal value
        Args:
            value (float): Nominal value to multiply with bounds
            norm (bool): If the input argument (bounds) are normalized, defaults to True
        """
        self._nominal = nominal
        self._normalize = norm
        return self

    def __eq__(self, other):

        if isinstance(other, dict):
            # if this is a dict, piece wise linear functions are being passed
            # Take the example of expenditure and capacity being modeled
            # sat the input is {(0, 200): 5000, (200, 300): 4000, (300, 400): 3000
            # what this implies is that for the capacity between 0 and 200, the expenditure is 5000
            # for the capacity between 200 and 300, the expenditure is 4000
            # for the capacity between 300 and 400, the expenditure is 3000
            # Modes objects index the bin. Three in this case : 1 - (0,200), 2 - (200,300), 3 - (300,400)
            # We need the following equations:
            # 1. capacity = capacity(bin0) + capacity(bin1) + capacity(bin2)
            # 2. x_capacity = 0*capacity(bin0) + 200*capacity(bin1) + 300*capacity(bin2), where is a reporting binary
            # 3-5. spend(bin0) = 5000*capacity(bin0); spend(bin1) = 4000*capacity(bin1); spend(bin2) = 3000*capacity(bin2)
            # 6. spend = spend(bin0) + spend(bin1) + spend(bin2)

            # this takes care of 1 and 2
            _ = self.sample == dict(enumerate(other))
            # this takes care of 3-6

            # the new modes object would have just been added to the model
            modes = self.model.modes[-1]

            _ = self.sample(modes)[self.calculation(modes)] == list(other.values())

        else:

            if self._nominal:
                if self._normalize:
                    other = [
                        (
                            (self._nominal * i[0], self._nominal * i[1])
                            if isinstance(i, tuple)
                            else self._nominal * i
                        )
                        for i in normalize(other)
                    ]
                else:
                    other = [
                        (
                            (self._nominal * i[0], self._nominal * i[1])
                            if isinstance(i, tuple)
                            else self._nominal * i
                        )
                        for i in other
                    ]

            if isinstance(other, Value):
                time = other.periods
                other = other.value

            else:
                time = None

            if self.sample.aspect not in self.model.dispositions:
                # if a calculation is given directly, without an explicit bound being set
                _ = self.sample == True

            if self._forall:
                if isinstance(other, list):
                    # if other is a list, iterate over the _forall indices
                    for n, idx in enumerate(self._forall):
                        _ = self(idx) == other[n]
                else:
                    for idx in self._forall:
                        # if other is not a list, just compare with the first element
                        _ = self(idx) == other

            else:
                # the aspect being calculated
                if time:
                    calc: Sample = self.calculation(time)
                else:
                    calc: Sample = self.calculation

                if self.sample.domain.modes:
                    # mode calculations, should map to modes
                    calc = calc(self.sample.domain.modes)

                # the aspect the calculation is dependant on
                decision: Sample = self.sample
                if self.sample.report:
                    # incidental, v_inc_calc = P_inc*x_v
                    v_lhs = calc.Vinc(other)
                    domain = calc.domain
                    v_rhs = decision.X(other)
                    cons_name = (
                        rf"{self.calculation.aspect.name}_inc{domain.idxname}_calc"
                    )

                else:
                    # v_calc = P*v
                    v_lhs = calc.V(other)
                    domain = calc.domain
                    v_rhs = decision.V(other)
                    cons_name = rf"{self.calculation.aspect.name}{domain.idxname}_calc"

                cons: C = v_lhs == other * v_rhs

                # categorize the constraint
                cons.categorize("Calculation")

                setattr(
                    self.program,
                    cons_name,
                    cons,
                )

                calc.aspect.constraints.append(cons_name)
                domain.update_cons(cons_name)

    def __call__(self, *index) -> Self:
        return Calculate(
            calculation=self.calculation(*index),
            sample=self.sample(*index),
        )



"""Library of elements"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ..components.commodity.currency import Currency
from ..components.impact.categories import Environ
from ..components.measure.unit import Unit
from ..components.temporal.scales import TemporalScales

if TYPE_CHECKING:
    from ..represent.model import Model


def time_units(model: Model):
    """Time units"""
    model.h = TemporalScales([1, 365, 24, 60, 60], ["y", "d", "h", "min", "s"])


def si_units(model: Model):
    """SI units"""
    # Base units
    model.m = Unit(label="m")  # meter
    model.kg = Unit(label="kg")  # kilogram

    model.km = 1000 * model.m  # kilometer

    # Derived units
    model.kW = 1000 * Unit(label="W")  # kilowatt
    model.MW = 1000 * model.kW  # megawatt
    model.GW = 1000 * model.MW  # gigawatt

    model.J = Unit(label="J")  # joule
    model.kJ = 1000 * model.J  # kilojoule
    model.MJ = 1000 * model.kJ  # megajoule

    model.siunits_set = True


def misc_units(model: Model):
    """Miscellaneous units"""

    if not model.siunits_set:
        si_units(model)

    model.L = 0.001 * Unit(label="m^3")  # liter
    model.gallon = 3.78541 * model.L  # gallon
    model.barrel = 158.987 * model.L  # barrel
    model.mmbtu = 1.055e9 * model.J  # million british thermal unit
    model.ton = 1000 * model.kg  # tonne
    model.mile = 1.60934 * Unit(label="mile")  # miles
    model.PJ = 1e15 * model.J  # petajoule
    model.GJ = 1e9 * model.J  # gigajoule


def env_indicators(model):
    """Load default indicators"""
    model.GWP = Environ(label="Global Warming Potential (kg CO2)")
    model.ODP = Environ(label="Ozone Depletion Potential (CFC-11 eq")
    model.AP = Environ(label="Acidification Potential (kg SO2)")
    model.EP = Environ(label="Eutrophication Potential (kg PO4)")
    model.POCP = Environ(label="Photochemical Ozone Creation Potential (kg C2H4)")
    model.HTP = Environ(label="Human Toxicity Potential (kg 1,4-DCB)")
    model.ETP = Environ(label="Ecotoxicity Potential (kg 1,4-DCB)")
    model.FFDP = Environ(label="Fossil Fuel Depletion Potential (kg oil)")
    model.MRDP = Environ(label="Mineral Resource Depletion Potential (kg resource)")


def currencies(model: Model):
    """Load default currencies

    citation
    --------
    Triennial Central Bank Survey: Foreign Exchange Turnover in April 2022.
    Bank for International Settlements,
    27 Oct. 2022, p. 12. Bank for International Settlements,
    https://www.bis.org/statistics/rpfx22_fx.pdf

    """
    model.USD = Currency(label="U.S. Dollar")
    model.EUR = Currency(label="Euro")
    model.JPY = Currency(label="Japanese Yen")
    model.GBP = Currency(label="Pound Sterling")
    model.CNY = Currency(label="Renminbi")
    model.AUD = Currency(label="Australian Dollar")
    model.CAD = Currency(label="Canadian Dollar")
    model.CHF = Currency(label="Swiss Franc")
    model.HKD = Currency(label="Hong Kong Dollar")
    model.SGD = Currency(label="Singapore Dollar")
    model.SEK = Currency(label="Swedish Krona")
    model.KRW = Currency(label="South Korean Won")
    model.NOK = Currency(label="Norwegian Krone")
    model.NZD = Currency(label="New Zealand Dollar")
    model.INR = Currency(label="Indian Rupee")
    model.MXN = Currency(label="Mexican Peso")
    model.TWD = Currency(label="New Taiwan Dollar")
    model.ZAR = Currency(label="South African Rand")
    model.BRL = Currency(label="Brazilian Real")
    model.DKK = Currency(label="Danish Krone")
    model.PLN = Currency(label="Polish Złoty")
    model.THB = Currency(label="Thai Baht")
    model.ILS = Currency(label="Israeli New Shekel")
    model.IDR = Currency(label="Indonesian Rupiah")
    model.CZK = Currency(label="Czech Koruna")
    model.AED = Currency(label="UAE Dirham")
    model.TRY_ = Currency(label="Turkish Lira")  # TRY is reserved in Python
    model.HUF = Currency(label="Hungarian Forint")
    model.CLP = Currency(label="Chilean Peso")
    model.SAR = Currency(label="Saudi Riyal")
    model.PHP = Currency(label="Philippine Peso")
    model.MYR = Currency(label="Malaysian Ringgit")
    model.COP = Currency(label="Colombian Peso")
    model.RUB = Currency(label="Russian Ruble")
    model.RON = Currency(label="Romanian Leu")
    model.PEN = Currency(label="Peruvian Sol")



"""Process conversion models (external libraries)"""

import numpy as np
from pandas import DataFrame

try:
    from pvlib.location import Location as PVLocation
    from pvlib.modelchain import ModelChain as PVModelChain
    from pvlib.pvsystem import PVSystem, retrieve_sam
    from pvlib.temperature import TEMPERATURE_MODEL_PARAMETERS

    import_all = False
except ImportError:
    import_all = True

try:
    from windpowerlib import ModelChain as WModelChain
    from windpowerlib import WindTurbine

    import_all = False
except ImportError:
    import_all = True


def pv(
    data: DataFrame,
    coord: tuple,
    sam: str = "cecmod",
    module_params: str = "Canadian_Solar_Inc__CS5P_220M",
    inverter: str = "cecinverter",
    inverter_params: str = "ABB__MICRO_0_25_I_OUTD_US_208__208V_",
    temperature_params: str = "open_rack_glass_glass",
    aoi_model: str = "no_loss",
    ac_model: str = "sandia",
    spectral_model: str = "no_loss",
):
    """Calculates solar power output using weather data
    Relevant factors include DHI (W/m2), DNI (W/m2), GHI (W/m2),
    Temperature (C), Dewpoint (C), Relative humidity (%)

    Args:
        data (DataFrame): weather data input with dni, dhi, wind_speed, ghi, air_temperature, dew_point, relative_humidity
        coord (tuple): latitude and longitude
        sam (str, optional): Defaults to 'cecmod'.
        module_parameters (str, optional): Defaults to 'Canadian_Solar_Inc__CS5P_220M'.
        inverter (str, optional):  Defaults to 'cecinverter'.
        inverter_parameters (str, optional): Defaults to 'ABB__MICRO_0_25_I_OUTD_US_208__208V_'.
        temperature_params (str, optional): Defaults to 'open_rack_glass_glass'.
        aoi_model (str, optional): Defaults to 'no_loss'.
        ac_model (str, optional): Defaults to 'sandia'.
        spectral_model (str, optional): Defaults to 'no_loss'.

    Returns:
        output (DataFrame): a dataframe with hourly solar power outputs
    """
    # data = data.resample('H').mean()

    if import_all:
        print(
            "This is an optional feature. Please install pvlib, or pip install energiapy[all]",
        )
        return None

    modules = retrieve_sam(sam)
    module_parameters = modules[module_params]
    inverters = retrieve_sam(inverter)
    inverter_parameters = inverters[inverter_params]
    tparams = TEMPERATURE_MODEL_PARAMETERS["sapm"][temperature_params]
    system = PVSystem(
        module_parameters=module_parameters,
        inverter_parameters=inverter_parameters,
        temperature_model_parameters=tparams,
    )
    location = PVLocation(latitude=coord[0], longitude=coord[1])
    mc = PVModelChain(
        system,
        location,
        aoi_model=aoi_model,
        ac_model=ac_model,
        spectral_model=spectral_model,
    )
    mc.run_model(weather=data)
    # make a list, avoid np.float64
    array = np.maximum(0, np.nan_to_num(np.array(mc.results.ac)))
    return [float(i) for i in array]


def wf(
    data: DataFrame,
    roughness_length: float = 0.1,
    turbine_type: str = "V100/1800",
    hub_height: float = 92,
    wind_speed_model: str = "logarithmic",
    density_model: str = "ideal_gas",
    temperature_model: str = "linear_gradient",
    power_output_model: str = "power_coefficient_curve",
    density_correction: bool = True,
    obstacle_height: float = 0,
    observation_height: float = 10,
):
    """Calculates wind power output using weather data
    Relevant factors include wind speeds (m/s), temperature (K), and pressure(Pa)

    Args:
        data (DataFrame): weather data input with wind_speed, air_temperature, surface_pressure
        roughness_length (float, optional): Defaults to 0.1.
        turbine_type (str, optional): Defaults to 'V100/1800'.
        hub_height (float, optional): Defaults to 92.
        wind_speed_model (str, optional): Defaults to 'logarithmic'.
        density_model (str, optional): Defaults to 'ideal_gas'.
        temperature_model (str, optional): Defaults to 'linear_gradient'.
        power_output_model (str, optional): Defaults to power_coefficient_curve.
        density_correction (bool, optional): Defaults to True.
        obstacle_height (float, optional): Defaults to 0.
        observation_height (float, optional): Defaults to 10.

    Returns:
        output (DataFrame): a dataframe with hourly wind power outputs
    """

    # df_ = df_.dropna()
    if import_all:
        print(
            "This is an optional feature. Please install windpowerlib, or pip install energiapy[all]",
        )
        return None
    data["pressure"] = 100 * data["surface_pressure"]
    data["temperature"] = data["air_temperature"] + 273.15
    data["roughness_length"] = roughness_length

    # data = data.resample('H').mean()
    data = DataFrame(
        data[["wind_speed", "temperature", "pressure", "roughness_length"]].to_numpy(),
        index=data.index,
        columns=[
            np.array(["wind_speed", "temperature", "pressure", "roughness_length"]),
            np.array([observation_height, observation_height, observation_height, 0]),
        ],
    )
    # specification of wind turbine where power curve is provided in the
    # oedb turbine library
    turbine_type_ = {
        "turbine_type": turbine_type,  # Vestas
        "hub_height": hub_height,  # in m
    }
    # initialize WindTurbine object
    turbine = WindTurbine(**turbine_type_)

    modelchain_data = {
        "wind_speed_model": wind_speed_model,  # 'logarithmic' (default),
        # 'hellman' or
        # 'interpolation_extrapolation'
        "density_model": density_model,  # 'barometric' (default), 'ideal_gas'
        # or 'interpolation_extrapolation'
        "temperature_model": temperature_model,  # 'linear_gradient' (def.) or
        # 'interpolation_extrapolation'
        # 'power_curve' (default) or
        "power_output_model": power_output_model,
        # 'power_coefficient_curve'
        "density_correction": density_correction,  # False (default) or True
        "obstacle_height": obstacle_height,  # default: 0
        "hellman_exp": None,
    }  # None (default) or None
    # initialize ModelChain with own specifications and use run_model method to
    # calculate power output
    mc_turbine = WModelChain(turbine, **modelchain_data).run_model(data)
    return list(mc_turbine.power_output)



"""Process"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional
from warnings import warn

from ...modeling.parameters.conversion import Conversion
from ._operation import _Operation

if TYPE_CHECKING:
    from ..commodity.resource import Resource
    from ..spatial.location import Location
    from ..temporal.lag import Lag
    from ..temporal.periods import Periods
    from .storage import Storage


@dataclass
class Process(_Operation):
    """Process converts one Resource to another Resource
    at some Location

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conv: Operational conversion associated with the operation. Defaults to None.
    :vartype conv: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool
    :ivar locations: Locations at which the process is balanced. Defaults to [].
    :vartype locations: list[Location]
    :ivar ofstorage: If the Process is Storage charging and discharging. Defaults to None.
    :vartype ofstorage: Storage, optional
    """

    def __post_init__(self):
        _Operation.__post_init__(self)

        # at which locations the process is balanced
        # Note that we do not need a conversion at every temporal scale.
        # once balanced at a location for a particular time,
        # if time != horizon, the individual streams are summed up anyway
        self.locations: list[Location] = []

        self.ofstorage: Optional[Storage] = None

    @property
    def spaces(self) -> list[Location]:
        """Locations at which the process is balanced"""
        return self.locations

    def writecons_conversion(self, loc_times: list[tuple[Location, Periods]]):
        """Write the conversion constraints for the process"""

        def time_checker(res: Resource, loc: Location, time: Periods):
            """This checks if it is actually necessary
            to write conversion at denser temporal scales
            """
            # This checks whether some other aspect is defined at
            # a lower temporal scale

            if loc not in self.model.grb[res]:
                # if not defined for that location, check for a lower order location
                # i.e. location at a lower hierarchy,
                # e.g. say if loc being passed is a city, and a grb has not been defined for it
                # then we need to check at a higher order
                parent = self.space.split(loc)[1]  # get location at one hierarchy above
                if parent:
                    # if that indeed exists, then make the parent the loc
                    # the conversion Balance variables will feature in grb for parent location
                    loc = parent

            _ = self.model.grb[res][loc][time]

            if res.inv_of:
                # for inventoried resources, the conversion is written
                # using the time of the base resource's grb
                res = res.inv_of

            try:
                times = list(
                    [
                        t
                        for t in self.model.grb[res][loc]
                        if self.model.grb[res][loc][t]
                    ],
                )
            except KeyError:
                times = []
            # write the conversion balance at
            # densest temporal scale in that space
            if times:
                return min(times)

            return time.horizon

        if not self.conv:
            warn(
                f"{self}: Conversion not defined, no Constraints generated",
                UserWarning,
            )
            return

        # This makes the conversion consistent
        # check conv_test.py in tests for examples
        self.conv.balancer()

        if self.conv.pwl:
            # if there are piece-wise linear conversions
            # here we assume that the same resources appear in all piece-wise segments
            # this is a reasonable assumption for conversion in processes
            # but not if process modes involve different resources

            # TODO:
            # make the statement eff = [conv[res] for conv in self.conversion.values()]
            # into try
            # if that fails, create a consistent dict, see:
            # {0: {r1: 10, r2: -5}, 1: {r1: 8, r2: -4, r3: -2}}
            # transforms to {0: {r1: 10, r2: -5, r3: 0}, 1: {r1: 8, r2: -4, r3: -2}}
            # the r3: 0 will ensure that r3 is considered in all modes
            # the zero checks will prevent unnecessary constraints
            # there is a problem though, because I am only checking for the elements in the first dict
            # in the multi conversion dict

            conversion = self.conversion[list(self.conversion)[0]]

        else:
            conversion = self.conversion

        for location, time in loc_times:

            if location in self.locations:
                # if the process is already balanced for the location , Skip
                continue

            for res, par in conversion.items():
                # set, the conversion on the resource

                setattr(res, self.name, self)
                # now there are two cases possible
                # the parameter (par) is positive or negative
                # if positive, the resource is expended
                # if negative, the resource is produced
                # also, the par can be an number or a list of numbers

                # insitu resource (produced and expended within the system)
                # do not initiate a grb so we need to run a check for that first

                if res in self.model.grb:
                    time = time_checker(res, location, time)
                    _ = self.model.grb[res].get(location, {})

                eff = par if isinstance(par, list) else [par]

                if eff[0] < 0:
                    # Resources are consumed (expendend by Process) immediately
                    rhs = res.expend(self.operate, location, time)
                    eff = [-e for e in eff]
                    opr = self.operate(location, time)
                else:
                    # Production — may occur after lag
                    lag_time = self.lag.of if self.lag else time
                    rhs = res.produce(self.operate, location, lag_time)
                    opr = self.operate(location, time)

                # because of using .balancer(), expend/produce are on same temporal scale

                if self.conv.pwl:

                    eff = [conv[res] for conv in self.conversion.values()]

                    if eff[0] < 0:
                        eff = [-e for e in eff]

                    if not self.conv.modes_set:
                        # this is setting the bin limits for piece wise linear conversion
                        # these are written bound to capacity generally
                        # but here we pause that binding and bind operate to explicit limits
                        self.model.operate.bound = None

                        _ = opr == dict(enumerate(self.conversion.keys()))

                        # reset capacity binding
                        self.model.operate.bound = self.model.capacity

                        modes = self.model.modes[-1]
                        self.conv.modes_set = True

                    else:
                        modes = self.conv.modes
                        modes.bind = self.operate
                        self.conv.modes_set = True

                    opr = opr(modes)

                    rhs = rhs(modes)

                    # _ = opr(modes)[rhs(modes)] == eff

                # else:
                _ = opr[rhs] == eff

            # update the locations at which the process exists
            self.locations.append(location)

    def __call__(
        self, resource: Resource | Conversion, lag: Optional[Lag] = None
    ) -> Conversion:
        """Conversion is called with a Resource to be converted"""

        if not self._conv:
            self.conv = Conversion(operation=self)
            self._conv = True

        if lag:
            return self.conv(resource, lag)
        return self.conv(resource)



"""Loss Variable"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from .aspect import Aspect

if TYPE_CHECKING:
    from ..indices.domain import Domain


@dataclass
class Loss(Aspect):
    """Loss Entails the degradation of a state
    In the case of Capacity, it is the loss of capacity
    In the case of Inventory, it is the loss of inventory
    Unlike Dismantle, loss is not a decision
    """

    def update(self, domain: Domain):
        """Add a domain to the decision variable"""
        self.domains.append(domain)



"""State Variable"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional

from ..constraints.balance import Balance
from ..constraints.vmap import Map
from .aspect import Aspect

if TYPE_CHECKING:
    from ..indices.domain import Domain
    from .control import Control


@dataclass
class State(Aspect):
    """State Variable
    Operational capacity or set point (utilization)
    """

    add: Optional[Control] = None
    sub: Optional[Control] = None
    bound: Optional[State] = None

    def __post_init__(self):
        Aspect.__post_init__(self)
        self.mapped_from: list[Domain] = []

    def update(self, domain: Domain, reporting: bool = False):
        """Add a domain to the decision variable"""

        # Write a mapping constraint
        Map(aspect=self, domain=domain, reporting=reporting)


@dataclass
class Size(State):
    """Capacity State Variable
    Operational capacity
    """

    def __post_init__(self):
        State.__post_init__(self)


@dataclass
class SetPoint(State):
    """Set Point State Variable
    Operational set point (utilization)
    """

    def __post_init__(self):
        State.__post_init__(self)


@dataclass
class Impact(State):
    """Consequence of an action"""

    def __post_init__(self):
        State.__post_init__(self)


@dataclass
class Stream(State):
    """Stream of Resource, also a state variable"""

    def __post_init__(self):
        State.__post_init__(self)

    # def update(self, domain: Domain):
    #     """Add a domain to the decision variable"""

    #     # Write a mapping constraint

    #     Map(aspect=self, domain=domain)

    #     # Write/Update, the stream balance
    #     Balance(aspect=self, domain=domain)


@dataclass
class EndoStream(Stream):
    """Endogenous Stream of Resource, also a state variable"""

    def __post_init__(self):
        Stream.__post_init__(self)

    def update(self, domain: Domain, reporting: bool = False):

        Map(aspect=self, domain=domain, reporting=reporting)

        Balance(aspect=self, domain=domain)


@dataclass
class ExoStream(Stream):
    """Exogenous Stream of Resource, also a state variable"""

    def __post_init__(self):
        Stream.__post_init__(self)

    def update(self, domain: Domain, reporting: bool = False):

        Map(aspect=self, domain=domain, reporting=reporting)

        Balance(aspect=self, domain=domain)


@dataclass
class IndStream(Stream):
    """Indicator Stream of Resource, also a state variable"""

    def __post_init__(self):
        Stream.__post_init__(self)

    def update(self, domain: Domain, reporting: bool = False):

        Map(aspect=self, domain=domain, reporting=reporting)



"""Temporal Scales"""

# from itertools import accumulate
# from operator import mul


class TemporalScales:
    """
    A temporal scale for a model.

    :param discretizations: List of discretizations for the temporal scale.
    :type discretizations: list[int]
    :param names: Names of the discretizations. Defaults to [t<i>] for each discretization.
    :type names: list[str], optional
    """

    def __init__(self, discretizations: list[int], names: list[str] = None):

        # self.discretization_list = list(accumulate(discretization_list, mul))
        self.discretizations = discretizations
        self.names = names if names else [f"t{i}" for i in discretizations]



"""Energy system examples"""

from ...components.commodity.currency import Currency
from ...components.commodity.material import Material
from ...components.commodity.resource import Resource
from ...components.impact.categories import Environ
from ...components.measure.unit import Unit
from ...components.operation.process import Process
from ...components.operation.storage import Storage
from ...components.spatial.location import Location
from ...components.temporal.periods import Periods
from ...components.temporal.scales import TemporalScales
from ...represent.model import Model


def scheduling():
    """A small scheduling example"""
    m = Model("scheduling")

    # ## Time
    #
    # We have 4 quarter which form a year

    m.q = Periods()
    m.y = 4 * m.q

    # The horizon ($\overset{\ast}{t} = argmin_{n \in N} |\mathcal{T}_{n}|$) is determined by Energia implicitly. Check it by printing Model.Horizon

    m.horizon

    # ## Space
    #
    # If nothing is provided, a default location is created. The created single location will serve as the de facto network
    #
    # In single location case studies, it may be easier to skip providing a location all together

    m.network

    # ## Resources
    #
    # In the Resource Task Network (RTN) methodology all commodities are resources. In this case, we have a few general resources (wind, power) and a monetary resource (USD).

    m.usd = Currency()
    m.power, m.wind = Resource(), Resource()

    # ### Setting Bounds
    #
    # The first bound is set for over the network and year, given that no spatiotemporal disposition is provided:
    #
    # $\mathbf{cons}_{wind, network, year_0} \leq 400$
    #
    # For the second bound, given that a nominal is given, the list is treated as multiplicative factor. The length matches with the quarterly scale. Thus:
    #
    # $\mathbf{rlse}_{power, network, quarter_0} \geq 60$
    #
    # $\mathbf{rlse}_{power, network, quarter_1} \geq 70$
    #
    # $\mathbf{rlse}_{power, network, quarter_2} \geq 100$
    #
    # $\mathbf{rlse}_{power, network, quarter_3} \geq 30$
    #

    m.wind.consume <= 400
    m.power.release.prep(100) >= [0.6, 0.7, 1, 0.3]

    # Check the model anytime, using m.show(), or object.show()
    #
    # The first constraint is a general resource balance, generated for every resource at every spatiotemporal disposition at which an aspect regarding it is defined.
    #
    # Skip the True in .show() for a more concise set notation based print

    # ## Process

    # There are multiple ways to model this.
    #
    # Whats most important, however, is the resource balance. The resource in the brackets is the basis resource. For a negative basis, multiply the resource by the negative factor or just use negation if that applies.

    m.wf = Process()
    m.wf(m.power) == -1 * m.wind

    # Now set bounds on the extent to which the wind farm can operate. We are actually writing the following constraint:
    #
    # $\mathbf{opr} <= \phi \cdot \mathbf{cap}$
    #
    # However, we know the capacity, so it is treated as a parameter.
    #
    # Note that the incoming values are normalized by default. Use norm = False to avoid that

    m.wf.operate.prep(200, norm=False) <= [0.9, 0.8, 0.5, 0.7]

    # Add a cost to the process operation. This implies that the cost of operating is variable in every quarter. In general:
    #
    # $\dot{\mathbf{v}} == \theta \cdot \mathbf{v}$

    m.wf.operate[m.usd.spend] == [4000, 4200, 4300, 3900]

    # Note that printing can be achieved via the aspect [operate, spend, etc.] or the object

    # ## Locating the process
    #
    # The production streams are only generated once the process is places in some location. In this study, the only location is available is the default network.

    m.network.locate(m.wf)

    # Alternatively,

    # m.wf.locate(m.network)

    # # The Model
    #
    # The model consists of the following:
    #
    # 1. A general resource balances for wind in (network, year) and power in (network, quarter)
    # 2. Bounds on wind consumption [upper] and power release [lower], and wf operation [upper]
    # 3. Conversion constraints, giving produced and expended resources based on operation
    # 4. Calculation of spending USD in every quarter
    # 5. Mapping constraints for operate: q -> y and spend: q -> y (generated after objective is set)

    return m


def design_scheduling():
    """A small design and scheduling example"""
    m = Model("design_scheduling")
    m.q = Periods()
    m.y = 4 * m.q
    m.usd = Currency()
    m.declare(Resource, ["power", "wind", "solar"])
    m.solar.consume(m.q) <= 100
    m.wind.consume <= 400
    m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]

    m.wf = Process()
    m.wf(m.power) == -1 * m.wind
    m.wf.capacity.x <= 100
    m.wf.capacity.x >= 10
    m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    m.wf.capacity[m.usd.spend] == 990637 + 3354
    m.wf.operate[m.usd.spend] == 49

    m.pv = Process()
    m.pv(m.power) == -1 * m.solar
    m.pv.capacity.x <= 100
    m.pv.capacity.x >= 10
    m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]
    m.pv.capacity[m.usd.spend] == 567000 + 872046
    m.pv.operate[m.usd.spend] == 90000

    m.lii = Storage()
    m.lii(m.power) == 0.9
    m.lii.capacity.x <= 100
    m.lii.capacity.x >= 10
    m.lii.capacity[m.usd.spend] == 1302182 + 41432
    m.lii.inventory[m.usd.spend] == 2000

    m.network.locate(m.wf, m.pv, m.lii)

    return m


def design_scheduling_materials():
    """Design and scheduling considering materials and emissions from them"""
    m = Model("design_scheduling")
    m.scales = TemporalScales([1, 4], ["y", "q"])
    m.usd = Currency()
    m.gwp = Environ()

    m.declare(Resource, ["power", "wind", "solar"])
    m.solar.consume == True
    m.wind.consume == True
    m.power.release.prep(180) >= [0.6, 0.7, 0.8, 0.3]

    m.declare(
        Material,
        [
            "lir",
            "lib",
            "steel",
            "concrete",
            "glass",
            "si_mono",
            "si_poly",
        ],
    )

    m.lir.consume[m.gwp.emit] == 9600
    m.lib.consume[m.gwp.emit] == 2800
    m.steel.consume[m.gwp.emit] == 2121.152427
    m.steel.consume[m.usd.spend] == 670

    m.concrete.consume[m.gwp.emit] == 120.0378
    m.glass.consume[m.gwp.emit] == 1118.5
    m.si_mono.consume[m.gwp.emit] == 122239.1
    m.si_poly.consume[m.gwp.emit] == 98646.7

    m.wf = Process()
    m.wf.fab[0] == 109.9 * m.steel + 398.7 * m.concrete
    m.wf.fab[1] == 249.605 * m.steel + 12.4 * m.concrete
    m.wf(m.power) == {
        m.wf.fab.modes[0]: -2.857 * m.wind,
        m.wf.fab.modes[1]: -2.3255 * m.wind,
    }
    m.wf.capacity.x <= 100
    m.wf.capacity.x >= 10
    m.wf.operate.prep(norm=True) <= [0.9, 0.8, 0.5, 0.7]
    m.wf.capacity(m.wf.fab.modes)[m.usd.spend] == [1292000 + 29200, 3192734 + 101498]
    m.wf.operate[m.usd.spend] == 49

    m.pv = Process()
    m.pv.fab[0] == 70 * m.glass + 7 * m.si_mono
    m.pv.fab[1] == 70 * m.glass + 7 * m.si_poly
    m.pv(m.power) == {
        m.pv.fab.modes[0]: -5 * m.solar,
        m.pv.fab.modes[1]: -6.67 * m.solar,
    }
    m.pv.capacity.x <= 100
    m.pv.capacity.x >= 10
    m.pv.operate.prep(norm=True) <= [0.6, 0.8, 0.9, 0.7]
    m.pv.capacity[m.usd.spend] == 567000 + 872046
    m.pv.operate[m.usd.spend] == 90000

    m.lii = Storage()
    m.lii(m.power) == 0.9
    m.lii.fab[0] == 0.137 * m.lib + 1.165 * m.steel
    m.lii.fab[1] == 0.137 * m.lir + 1.165 * m.steel
    m.lii.capacity.x <= 100
    m.lii.capacity.x >= 10
    m.lii.capacity[m.usd.spend] == 1302182 + 41432
    m.lii.inventory[m.usd.spend] == 2000

    m.network.locate(m.wf, m.pv, m.lii)

    return m


def supermarket():
    """A supermarket energy system example

    Optimize for cost:
        m.usd.spend.opt()

    Optimize for CO2 emissions:
        m.co2_vent.release.opt()

    Maximize efficiency:
        m._.lb(sum(m._.consume))
    """
    resource_demand_dict = {
        "Lighting": 200,
        "Refrigeration": 1000,
        "Space Heating": 100,
    }

    resource_dict = {
        "Price": {"Natural Gas": 8.89, "Biomass": 9.72, "Grid Electricity": 36.11},
        "CO2 Generation": {"Natural Gas": 56, "Biomass": 100, "Grid Electricity": 90},
    }

    generation_process_dict = {
        "Biomass ST": {
            "nE": 68,
            "nH": 0,
            "LB": 100,
            "UB": 10**6,
            "Capex": 250,
            "Opex": 15,
        },
        "Natural Gas CHP": {
            "nE": 44,
            "nH": 28,
            "LB": 800,
            "UB": 10**6,
            "Capex": 500,
            "Opex": 15,
        },
        "Solar PV": {"nE": 9, "nH": 0, "LB": 10, "UB": 300, "Capex": 2000, "Opex": 500},
        "Wind Farm": {
            "nE": 22,
            "nH": 0,
            "LB": 10,
            "UB": 500,
            "Capex": 2000,
            "Opex": 1200,
        },
    }

    consumption_process_dict = {
        "Refrigeration": {"Efficiency": 300, "Capex": 80, "Opex": 85},
        "LED": {"Efficiency": 70, "Capex": 10, "Opex": 30},
        "Heating": {"Efficiency": 4, "Capex": 1, "Opex": 3},
    }

    m = Model("supermarket")

    m.supermarket = Location()

    m.GJ = Unit(label="Giga Joules")
    m.PJ = 10**6 * m.GJ
    m.kW = Unit(label="kilo Watts")
    m.tons = Unit(label="US Tons")

    # Resources can be declared along with thier attributes such as maximum allowed consumption, dischargeablity, base price, etc.

    m.usd = Currency(label="$")

    m.biomass = Resource(basis=m.GJ, label="Biomass")
    m.biomass.consume == True
    m.biomass.consume[m.usd.spend] == resource_dict["Price"]["Biomass"]

    m.gridpower = Resource(basis=m.GJ, label="Grid Electricity")
    m.gridpower.consume == True
    m.gridpower.consume[m.usd.spend] == resource_dict["Price"]["Grid Electricity"]

    m.ng = Resource(basis=m.GJ, label="Natural gas")
    m.ng.consume == True
    m.ng.consume[m.usd.spend] == resource_dict["Price"]["Natural Gas"]

    m.power = Resource(basis=m.kW, label="Electrical Power")
    m.heat = Resource(basis=m.kW, label="Heat Power")

    m.co2_vent = Resource(basis=m.tons, label="Carbon dioxide")
    m.co2_vent.release == True

    m.solar = Resource(basis=m.GJ, label="Solar energy")
    m.solar.consume == True
    m.wind = Resource(basis=m.GJ, label="Wind energy")
    m.wind.consume == True

    m.lighting = Resource(basis=m.kW, label="Lighting")
    m.lighting.release >= resource_demand_dict["Lighting"]

    m.refrigeration = Resource(basis=m.kW, label="Refrigeration")
    m.refrigeration.release >= resource_demand_dict["Refrigeration"]

    m.heating = Resource(basis=m.kW, label="Heating")
    m.heating.release >= resource_demand_dict["Space Heating"]

    m.st = Process(basis=m.PJ, label="Biomass ST")
    m.st(-m.biomass) == (
        277.78 * generation_process_dict["Biomass ST"]["nE"] / 100
    ) * m.power + m.co2_vent * resource_dict["CO2 Generation"]["Biomass"]
    # using x sets a binary, making the setting up of the process optional
    m.st.capacity.x >= generation_process_dict["Biomass ST"]["LB"]
    m.st.capacity.x <= generation_process_dict["Biomass ST"]["UB"]
    m.st.capacity[m.usd.spend] == generation_process_dict["Biomass ST"]["Capex"] * 0.05
    m.st.operate[m.usd.spend] == generation_process_dict["Biomass ST"]["Opex"]

    m.chp = Process(basis=m.PJ, label="Biomass ST")
    m.chp(-m.ng) == (
        277.78 * generation_process_dict["Natural Gas CHP"]["nE"] / 100
    ) * m.power + (
        277.78 * generation_process_dict["Natural Gas CHP"]["nH"] / 100
    ) * m.heat + m.co2_vent * resource_dict[
        "CO2 Generation"
    ][
        "Natural Gas"
    ]
    # using x sets a binary, making the setting up of the process optional
    m.chp.capacity.x >= generation_process_dict["Natural Gas CHP"]["LB"]
    m.chp.capacity.x <= generation_process_dict["Natural Gas CHP"]["UB"]
    m.chp.capacity[m.usd.spend] == generation_process_dict["Natural Gas CHP"][
        "Capex"
    ] * 0.05
    m.chp.operate[m.usd.spend] == generation_process_dict["Natural Gas CHP"]["Opex"]

    m.pv = Process(basis=m.kW, label="Solar PV")
    m.pv(-m.solar) == (
        277.78 * generation_process_dict["Solar PV"]["nE"] / 100
    ) * m.power
    m.pv.capacity.x <= generation_process_dict["Solar PV"]["UB"]
    m.pv.capacity.x >= generation_process_dict["Solar PV"]["LB"]
    m.pv.capacity[m.usd.spend] == generation_process_dict["Solar PV"]["Capex"] * 0.05
    m.pv.operate[m.usd.spend] == generation_process_dict["Solar PV"]["Opex"]

    m.wf = Process(basis=m.kW, label="Wind Farm")
    m.wf(-m.wind) == (
        277.78 * generation_process_dict["Wind Farm"]["nE"] / 100
    ) * m.power
    m.wf.capacity.x <= generation_process_dict["Wind Farm"]["UB"]
    m.wf.capacity.x >= generation_process_dict["Wind Farm"]["LB"]
    m.wf.capacity[m.usd.spend] == generation_process_dict["Wind Farm"]["Capex"] * 0.05
    m.wf.operate[m.usd.spend] == generation_process_dict["Wind Farm"]["Opex"]

    m.grid = Process(basis=m.PJ, label="Grid Electricity")
    m.grid(-m.gridpower) == 277.78 * m.power + m.co2_vent * resource_dict[
        "CO2 Generation"
    ]["Grid Electricity"]
    m.grid.capacity <= 10**5  # no binary needed because no upper bound

    m.refrigerator = Process(basis=m.kW, label="Refrigerator")
    m.refrigerator.capacity <= 10**5
    m.refrigerator(-m.power) == (
        consumption_process_dict["Refrigeration"]["Efficiency"] / 100
    ) * m.refrigeration
    m.refrigerator.capacity[m.usd.spend] == consumption_process_dict["Refrigeration"][
        "Capex"
    ] * 0.05
    m.refrigerator.operate[m.usd.spend] == consumption_process_dict["Refrigeration"][
        "Opex"
    ]

    m.led = Process(basis=m.kW, label="LED")
    m.led(-m.power) == (
        consumption_process_dict["LED"]["Efficiency"] / 100
    ) * m.lighting
    m.led.capacity[m.usd.spend] == consumption_process_dict["LED"]["Capex"] * 0.05
    m.led.operate[m.usd.spend] == consumption_process_dict["LED"]["Opex"]

    m.heater = Process(basis=m.kW, label="Heater")
    m.heater(-m.heat) == (
        consumption_process_dict["Heating"]["Efficiency"] / 100
    ) * m.heating
    m.heater.capacity[m.usd.spend] == consumption_process_dict["Heating"][
        "Capex"
    ] * 0.05
    m.heater.operate[m.usd.spend] == consumption_process_dict["Heating"]["Opex"]

    # Locations serve as aggregations of scenarios

    m.supermarket.locate(
        m.st,
        m.chp,
        m.pv,
        m.wf,
        m.grid,
        m.refrigerator,
        m.led,
        m.heater,
    )

    # Various constraints are need for the formulating the mathematical programming model. Here we include the cost, production, resource balance, demand, inventory, and network constraints.

    return m



"""Graph"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...components.graph.edge import Edge
from ...components.graph.node import Node

if TYPE_CHECKING:
    from ..model import Model


@dataclass
class Graph:
    """Graph representation

    Attributes:
        model (Model): Model to which the graph belongs.
        name (str): Name of the graph. Defaults to None.
        nodes (list[Node]): List of nodes in the graph.
        edges (list[Edge]): List of edges in the graph.

    Note:
        - name is generated based on Model name
        - nodes and edges are populated as model is defined
    """

    model: Model

    def __post_init__(self):
        self.name = f"Graph({self.model})"
        self.nodes = []
        self.edges = []

    def __setattr__(self, name, value):
        # give a name to the Node or Edge
        # set self as the graph of the Node or Edge
        # add the Node or Edge to the graph

        if isinstance(value, Node):
            value.name = name
            value.graph = self
            self.nodes.append(value)
        elif isinstance(value, Edge):
            value.name = name
            value.graph = self
            self.edges.append(value)

        super().__setattr__(name, value)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""A Model"""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Optional, Self, Type

from dill import dump
from ppopt.mplp_program import MPLP_Program
from ppopt.solution import Solution as MPSolution

from .._core._x import _X
from ..components.commodity.currency import Currency
from ..components.commodity.emission import Emission
from ..components.commodity.land import Land
from ..components.commodity.material import Material
from ..components.commodity.resource import Resource
from ..components.game.couple import Couple
from ..components.game.player import Player
from ..components.impact.categories import Economic, Environ, Social
from ..components.measure.unit import Unit
from ..components.operation.process import Process
from ..components.operation.storage import Storage
from ..components.operation.transport import Transport
from ..components.spatial.linkage import Linkage
from ..components.spatial.location import Location
from ..components.temporal.modes import Modes
from ..components.temporal.periods import Periods
from ..components.temporal.scales import TemporalScales
from ..dimensions.consequence import Consequence
from ..dimensions.problem import Problem
from ..dimensions.space import Space
from ..dimensions.system import System
from ..dimensions.time import Time
from ..library.recipes import (
    capacity_sizing,
    economic,
    environmental,
    free_movement,
    inventory_sizing,
    operating,
    social,
    trade,
    usage,
)
from ..modeling.parameters.conversion import Conversion
from ..modeling.variables.control import Control
from ..modeling.variables.recipe import Recipe
from ..modeling.variables.states import Impact, State, Stream
from .ations.graph import Graph
from .ations.program import Program

if TYPE_CHECKING:
    from enum import Enum
    from typing import DefaultDict

    from gana.block.solution import Solution
    from gurobipy import Model as GPModel

    from .._core._component import _Component
    from ..components.commodity._commodity import _Commodity
    from ..modeling.indices.domain import Domain
    from ..modeling.variables.aspect import Aspect
    from ..modeling.variables.sample import Sample

    GRBType = DefaultDict[
        _Commodity,
        DefaultDict[Location | Linkage, DefaultDict[Periods, list[Aspect]]],
    ]


@dataclass
class Model:
    """
    An abstract representation of an energy system.

    :param name: Name of the Model. Defaults to m.
    :type name: str
    :param init: List of functions to initialize the Model. Defaults to None.
    :type init: list[Callable], optional
    :param default: True if some default objects should be declared.
    :type default: bool
    :param capacitate: True if process capacities should be determined to bound operations.
    :type capacitate: bool

    :ivar added: List of added objects to the Model.
    :vartype added: list[str]
    :ivar update_map: maps component type to representation and collection.
    :vartype update_map: dict
    :ivar time: time representation of the Model.
    :vartype time: Time
    :ivar space: spatial representation of the Model.
    :vartype space: Space
    :ivar impact: impact representation of the Model.
    :vartype impact: Impact
    :ivar tree: feasible region (Decision-Making) of the Model.
    :vartype tree: Tree
    :ivar graph: Graph (Network) of the Model.
    :vartype graph: Graph
    :ivar system: System (Resource Task Network) of the Model.
    :vartype system: System
    :ivar program: Mathematical (mixed integer) program of the Model.
    :vartype program: Program
    :ivar conversions: List of Balances in the Model.
    :vartype conversions: list[Conversion]
    :ivar convmatrix: Conversion matrix of the Model.
    :vartype convmatrix: dict[Process, dict[Resource, int | float | list]]
    :ivar modes_dict: Dictionary mapping Bind objects to Modes.
    :vartype modes_dict: dict[Bind, Modes]
    :siunits_set: True if SI units have been set.
    :vartype siunits_set: bool
    :ivar cookbook: Recipes to create Aspects.
    :vartype cookbook: dict[str, Recipe]
    :ivar attr_map: Map of attribute names to recipes for creating them.
    :vartype attr_map: dict[str, dict[str, Recipe]]
    :ivar classifiers: List of classifiers for the Model.
    :vartype classifiers: list[Enum]
    :ivar grb: Dictionary which tells you what aspects of resource have GRB {loc: time: []} and {time: loc: []}.
    :vartype grb: DefaultDict[_Commodity,DefaultDict[Location | Linkage, DefaultDict[Periods, list[Aspect]]]]
    :ivar dispositions: Dictionary which tells you what aspects of what component have been bound at what location and time.
    :vartype dispositions: dict[Aspect, dict[_Commodity | Process | Storage | Transport, dict[Location | Linkage, dict[Periods, list[Aspect]]]]]
    :ivar maps: Maps of aspects to domains.
    :vartype maps: dict[Aspect, dict[Domain, dict[str, list[Domain]]]]
    :ivar maps_report: Maps of aspects to domains for reporting variables.
    :vartype maps_report: dict[Aspect, dict[Domain, dict[str, list[Domain]]]]


    :raises ValueError: If an attribute name already exists in the Model.
    """

    name: str = "m"
    init: Optional[list[Callable[Self]]] = None
    default: bool = True
    capacitate: bool = False

    def __post_init__(self):

        # what components have been added to the model
        self.added: list[str] = []
        # map of what representation and collection within that representation
        # an object of a particular type belongs to

        # the structure of components:
        # I Temporal representation (Time):
        # 1.  Periods (Periods) generates a bespoke discretization.
        # II Spatial representation (Space):
        # 1. Spatial representation (Space). Location (Loc) generate a bespoke discretization.
        # III Streams (System):
        # 1. Commodity (Resource) of any kind
        # 2. Emission (Emission) resource
        # 3. Land (Land) resource
        # 4. Money (Currency)
        # 5. Material (Material) used to setup processes
        # 6. etc. societal (Jobs), etc (Etc).
        # IV Operations (System):
        # 1. A production operation (Process) which describes a task in the system that involves conversion of resources
        # 2. A storage operation (Storage) which describes a task in the system that involves storing (charge) resources
        # and retrieving (discharge) them at later times.
        # 3. A transport operation (Transport) which describes a task in the system that involves transporting resources from
        #    one location to another.
        # V Impact, scales a stream and projects onto a common metric
        #   1. Impact (Impact) categories include Eco, Soc

        self.update_map = {
            Periods: ("time", "periods"),
            Modes: ("time", "modes"),
            Location: ("space", "locations"),
            Linkage: ("space", "linkages"),
            Environ: (
                "consequence",
                "envs",
            ),
            Social: (
                "consequence",
                "socs",
            ),
            Economic: ("consequence", "ecos"),
            Process: ("system", "processes"),
            Storage: ("system", "storages"),
            Transport: ("system", "transports"),
            Player: ("system", "players"),
            Couple: ("system", "couples"),
            Currency: ("system", "currencies"),
            Land: ("system", "lands"),
            Emission: ("system", "emissions"),
            Material: ("system", "materials"),
            Resource: ("system", "resources"),
            Control: ("problem", "controls"),
            Stream: ("problem", "streams"),
            State: ("problem", "states"),
            Impact: ("problem", "impacts"),
        }

        self.dimension_map = {
            collection: dimension for dimension, collection in self.update_map.values()
        }

        self.program_collections = [
            "constraint_sets",
            "function_sets",
            "variable_sets",
            "parameter_sets",
            "theta_sets",
            "index_sets",
            "constraints",
            "functions",
            "variables",
            "thetas",
            "indices",
            "objectives",
            "A",
            "B",
            "C",
            "F",
            "G",
            "H",
            "CrA",
            "CrB",
            "NN",
            "A_with_NN",
            "B_with_NN",
            "Z",
            "P",
        ]

        self.default_components = {
            "l": self.default_location,
            "t0": self.default_periods,
            "t": self.default_periods,
            "money": self.default_currency,
        }

        self.graph_components = ["edges", "nodes"]

        # map of attribute names to recipes for creating them
        self.attr_map: dict[str, dict[str, Recipe]] = {}

        # added attributes mapping to the created Aspect objects
        self._attr_map: dict[str, Aspect] = {}

        # maps to recipes for creating aspects
        self.cookbook: dict[str, Recipe] = {}

        # Temporal Scope
        self.time = Time(self)
        # Spatial Scope
        self.space = Space(self)

        # Impact on the exterior
        self.consequence = Consequence(self)

        # System (Resource Task Network)
        self.system = System(self)

        # Graph (Network)
        self.graph = Graph(self)

        # the problem
        self.problem = Problem(self)

        # mathematical program
        self.program = Program(model=self)
        # shorthand
        self._ = self.program

        # measuring units
        self.units: list[Unit] = []
        self.conversions: list[Conversion] = []  # not added to program
        self.convmatrix: dict[Process, dict[Resource, int | float | list]] = {}

        self.modes_dict: dict[Sample, Modes] = {}

        # if SI units have been set
        self.siunits_set: bool = False

        if not self.init:
            self.init = []

        if self.default:
            self.init += [
                capacity_sizing,
                operating,
                inventory_sizing,
                free_movement,
                trade,
                economic,
                environmental,
                social,
                usage,
            ]

        for func in self.init:
            func(self)

        # # introduce the dimensions of the model
        # Decisions.__post_init__(self)

        self.classifiers: dict[str, list[Enum]] = {
            "uncertainty": [],
            "structure": [],
            "scale": [],
            "paradigm": [],
        }

        # Dictionary which tells you what aspects of resource
        # have been set in what location and time
        self.grb: dict[
            _Commodity,
            dict[Location | Linkage, dict[Periods, list[Aspect]]],
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

        # Dictionary which tells you what aspects of what component
        # have been bound at what location and time
        self.dispositions: dict[
            Aspect,
            dict[
                _Commodity | Process | Storage | Transport,
                dict[Location | Linkage, dict[Periods, list[Aspect]]],
            ],
        ] = {}

        self.maps: dict[Aspect, dict[Domain, dict[str, list[Domain]]]] = {}
        self.maps_report: dict[Aspect, dict[Domain, dict[str, list[Domain]]]] = {}

    # -----------------------------------------------------
    #              Set Component
    # -----------------------------------------------------

    @property
    def horizon(self) -> Periods:
        """The horizon of the Model"""
        return self.time.horizon

    @property
    def network(self) -> Location:
        """The network of the Model"""
        return self.space.network

    @property
    def indicators(self) -> list[Social | Environ | Economic]:
        """Indicators"""
        return self.consequence.indicators

    @property
    def operations(self) -> list[Process | Storage | Transport]:
        """The Operations"""
        return self.system.operations

    @property
    def solution(self) -> dict[int, Solution | MPSolution]:
        """The solution of the program"""
        return self.program.solution

    @property
    def formulation(self) -> dict[int, GPModel | MPLP_Program]:
        """The formulations of the program"""
        return self.program.formulation

    def update(
        self,
        name: str,
        value: _X,
        represent: str,
        collection: str,
        aspects: Optional[list[str]] = None,
    ):
        """Update the Model with a new value

        Args:
            name (str): Name of the value to be added
            value (X): Value to be added
            represent (str): Representation to which the value belongs
            collection (str): Collection within the representation to which the value belongs
            subset (bool, optional): If True, the value is not added to the Model's
        """

        value.name = name
        # every component is handed the model
        value.model = self

        if name in self.added:
            # do not allow overriding of components
            # throw error if name already exists
            raise ValueError(f"{name} already defined")
            # added is the list of all components that have been added to the model
        self.added.append(name)

        # if not subset:
        #     # ignore subsets

        model_set: list = getattr(getattr(self, represent), collection)
        # the set that needs to be updated
        model_set.append(value)

        # update the index set for index elements
        if collection in [
            "resources",
            "currencies",
            "lands",
            "emissions",
            "materials",
            "processes",
            "storages",
            "transits",
            "locations",
            "linkages",
        ]:
            setattr(
                self.program, collection, getattr(self.program, collection) | value.I
            )

        # set aspects (as binds) on the components
        if aspects:
            for asp in aspects:
                aspect = getattr(self, asp)

                setattr(value, asp, aspect(value))

                if aspect.neg is not None:
                    setattr(value, aspect.neg.name, aspect.neg(value))

    def __setattr__(self, name, value):

        if isinstance(value, (str, dict, list, bool)) or value is None:
            # if value is a string, dict, list or bool
            # set the attribute to the value
            super().__setattr__(name, value)
            return

        if isinstance(value, TemporalScales):
            self.TemporalScales(value.discretizations, value.names)

            return

        if isinstance(value, Unit):
            value.name = name
            self.units.append(value)

        # map to representation and collection
        for cls, updates in self.update_map.items():
            if isinstance(value, cls):
                # for args in updates:
                self.update(name, value, *updates)
                break

        # Special linkage instructions
        if isinstance(value, Linkage):

            self.space.sources.append(value.source)
            self.space.sinks.append(value.sink)

            if value.bi:
                # if bidirectional, set the reverse linkage
                # also ensures that all linakges go in one direction only
                rev = value.rev()
                setattr(self, rev.name, rev)

        super().__setattr__(name, value)

    def __getattr__(self, name):

        if name in self.default_components:
            component = self.default_components[name]()
            return component

        if name in self.dimension_map:
            dimension = getattr(self, self.dimension_map[name])
            collection = getattr(dimension, name)
            setattr(self, name, collection)
            return collection

        if name in self.program_collections:
            collection = getattr(self.program, name)
            setattr(self, name, collection)
            return collection

        # Only called when attribute does not exist
        if name in self._attr_map:
            # if attribute has been called before
            # the next time the created attribute is returned
            return self._attr_map[name]

        if name in self.cookbook:
            recipe = self.cookbook[name]

            aspect = recipe.kind(**recipe.args)
            setattr(self, name, aspect)

            return aspect

        if name in self.attr_map:
            # if this is an attribute being called for the first time
            recipe = self.attr_map[name]
            aspect_name = list(recipe.keys())[0]

            if aspect_name in self.added:
                # if same aspect is called by a different name
                return getattr(self, aspect_name)

            # these are the arguments for the aspect
            recipe = recipe[aspect_name]
            aspect = recipe.kind(**recipe.args)

            setattr(self, aspect_name, aspect)

            self._attr_map[name] = aspect

            return aspect

        raise AttributeError(
            f"{self} has no '{name}'",
        )

    def aliases(self, *names: str, to: str):
        """Set aspect aliases"""
        _add = dict.fromkeys(list(names), {to: self.cookbook[to]})
        self.attr_map = {**self.attr_map, **_add}

    def Recipe(
        self,
        name: str,
        kind: Type[Aspect],
        primary_type: tuple[Type[_Component]] | Type[_Component],
        label: str = "",
        add: str = "",
        add_latex: str = "",
        add_kind: Type[Aspect] = None,
        sub: str = "",
        sub_latex: str = "",
        sub_kind: Type[Aspect] = None,
        neg: str = "",
        neg_latex: str = "",
        neg_label: str = "",
        bound: str = "",
        ispos: bool = True,
        nn: bool = True,
        latex: str = None,
    ):
        """Creates a Recipe and updates recipes

        Args:
            name (str): name of the aspect
            kind (Type[Aspect]): type of the aspect
            primary_type (tuple[Type[_Component]] | Type[_Component]): type of primary component
            label (str, optional): label for the aspect. Defaults to ''.
            add (str, optional): add control variable. Defaults to ''.
            sub (str, optional): sub control variable. Defaults to ''.
            neg (str, optional): name of the negative aspect. Defaults to ''.
            ispos (bool, optional): whether the aspect is positive. Defaults to True.
            nn (bool, optional): whether the aspect is non-negative. Defaults to True.
            latex (str, optional): LaTeX representation for the aspect. Defaults to None.
        """
        if name in self.cookbook:
            print(f"--- Warning: Overriding existing recipe ---{name}")

        self.cookbook[name] = Recipe(
            name=name,
            kind=kind,
            label=label or name,
            add=add,
            sub=sub,
            bound=bound,
            ispos=ispos,
            nn=nn,
            primary_type=primary_type,
            latex=latex,
        )

        if add:
            if not add_kind:
                if sub_kind:
                    add_kind = sub_kind
            self.Recipe(
                name=add,
                kind=add_kind,
                primary_type=primary_type,
                label=add_latex or add,
                ispos=True,
                nn=True,
                latex=latex or add,
            )

        if sub:
            if not sub_kind:
                if add_kind:
                    sub_kind = add_kind

            self.Recipe(
                name=sub,
                kind=sub_kind,
                primary_type=primary_type,
                label=sub_latex or sub,
                ispos=False,
                nn=True,
                latex=latex or sub,
            )

        if neg:
            neg_recipe = Recipe(
                name=neg,
                kind=kind,
                label=neg_label,
                ispos=not ispos,
                nn=nn,
                primary_type=primary_type,
                latex=neg_latex or neg,
            )
            self.cookbook[neg] = neg_recipe

    # -----------------------------------------------------
    #              Birth Component
    # -----------------------------------------------------

    def declare(self, what: Type[_X], names: list[str]):
        """Declares objects conveniently"""
        for i in names:
            setattr(self, i, what())

    def Link(
        self,
        source: Location,
        sink: Location,
        dist: float | Unit = None,
        bi: bool = False,
    ):
        """Link two Locations"""
        if source - sink:
            # if source and sink are already linked
            raise ValueError(
                f"A link already defined between {source} and {sink}.\n"
                "For multiple linkages with different attributes, use model.named_link = Link(...)",
            )

        link = Linkage(source=source, sink=sink, dist=dist, bi=bi, auto=True)
        setattr(self, f"{source.name}-{sink.name}", link)

    def TemporalScales(self, discretizations: list[int], names: list[str]):
        """
        This is an easy way to define multiple time periods (scales)

        :param discretizations: List of discretizations for the temporal scale.
        :type discretizations: list[int]
        :param names: Names of the discretizations. Defaults to [t<i>] for each discretization.
        :type names: list[str], optional
        """
        # set the root period:
        setattr(self, names[-1], Periods())
        # pick up the period that was just created
        # use it as the root
        root = self.periods[-1]
        discretizations = list(reversed(discretizations))

        names = list(reversed(names[:-1]))

        if discretizations[-1] != 1:
            discretizations.append(1)
            names.append("t0")

        for disc, name in zip(discretizations, names):
            setattr(self, name, disc * root)
            root = self.periods[-1]

    def show(
        self,
        descriptive: bool = False,
        categorical: bool = True,
        category: str = None,
    ):
        """Pretty print the Model"""
        self.program.show(descriptive, categorical=categorical, category=category)

    def sol(self, n_sol: int = 0, slack: bool = True, compare: bool = False):
        """Solution"""
        return self.program.sol(n_sol=n_sol, slack=slack, compare=compare)

    def eval(
        self, *theta_vals: float, n_sol: int = 0, roundoff: int = 4
    ) -> list[float]:
        """Evaluate the objective function at given theta values"""
        return self.program.eval(*theta_vals, n_sol=n_sol, roundoff=roundoff)

    def save(self, as_type: str = "dill"):
        """Save the Model to a file"""
        if as_type == "dill":
            with open(self.name + ".energia", "wb") as f:
                dump(self.solution, f)
        else:
            raise ValueError(f"Unknown type {as_type} for saving the model")

    def draw(self, variable: Aspect | Sample = None, n_sol: int = 0):
        """Draw the solution for a variable"""
        if variable is not None:
            self.program.draw(variable=variable.V(), n_sol=n_sol)

        else:
            self.program.draw(n_sol=n_sol)

    def default_periods(self, size: int = 0) -> Periods:
        """Return a default period"""

        if size:
            # if size is passed,
            # make a new temporal scale
            new_period = Periods(f"Time/{size}", periods=size, of=self.horizon)
            setattr(self, f"t{len(self.time.periods)}", new_period)

            # return the newly created period
            return self.time.periods[-1]

        # or create a default period

        self.t0 = Periods("Time")
        return self.t0

    def default_location(self) -> Location:
        """Return a default location"""
        self.l = Location(label="l")
        return self.l

    def default_currency(self) -> Currency:
        """Return a default currency"""
        if self.currencies:
            return self.currencies[0]

        self.money = Currency(label="$")
        return self.money

    def locate(self, *operations: Process | Storage):
        """Locate operations in the network"""
        self.network.locate(*operations)

    def solve(
        self,
        using: Literal[
            "combinatorial",
            "combinatorial_parallel",
            "combinatorial_parallel_exp",
            "graph",
            "graph_exp",
            "graph_parallel",
            "graph_parallel_exp",
            "combinatorial_graph",
            "geometric",
            "geometric_parallel",
            "geometric_parallel_exp",
        ] = "combinatorial",
    ):
        """Solve the multiparametric program"""

        self.program.solve(using=using)

    def __call__(self, *funcs: Callable[Self]):
        """Set functions on the model

        These can include default units

        """

        for f in funcs:
            f(self)

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



from dataclasses import dataclass

# from ..impact.categories import Eco
from .resource import Resource

# from ..operation.task import Task


@dataclass
class Package(Resource):
    """Package, discrete

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """


@dataclass
class Human(Resource):
    """Human

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """


@dataclass
class Mana(Resource):
    """Mana

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """


@dataclass
class Etc(Resource):
    """Etc, used for resources that do not fit into the other categories

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """



"""Bound Parameter Input"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._name import _Name

if TYPE_CHECKING:
    from ...components.temporal.periods import Periods


@dataclass
class Value(_Name):
    """Input Value"""

    value: (
        float
        | int
        | list[float]
        | list[int]
        | tuple[float]
        | tuple[int]
        | tuple[list[int | float]]
    ) = None
    periods: Periods = None

    def __post_init__(self):
        if isinstance(self.value, (float, int)):
            self.name = f"{self.value}/{self.periods}"
        else:
            self.name = f"Φ/{self.periods}"



"""Impact Indicator"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._component import _Component

if TYPE_CHECKING:
    from ...dimensions.consequence import Consequence


@dataclass
class Indicator(_Component):
    """Impact Indicator

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    """

    def __post_init__(self):
        _Component.__post_init__(self)

    @property
    def consequence(self) -> Consequence:
        """Impact object"""
        return self.model.consequence



"""Transport"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...modeling.parameters.conversion import Conversion
from ._operation import _Operation

if TYPE_CHECKING:
    from ..commodity.resource import Resource
    from ..spatial.linkage import Linkage
    from ..spatial.location import Location
    from ..temporal.lag import Lag
    from ..temporal.periods import Periods


@dataclass
class Transport(_Operation):
    """Exports Resource through Link
    basically, moves Resources between Locations

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conv: Operational conversion associated with the operation. Defaults to None.
    :vartype conv: Conversion, optional
    :ivar _conv: True if the operational conversion has been set. Defaults to False.
    :vartype _conv: bool
    :ivar fab: Material conversion associated with the operation. Defaults to None.
    :vartype fab: Conversion, optional
    :ivar _fab_balanced: True if the material conversion has been balanced. Defaults to False.
    :vartype _fab_balanced: bool
    :ivar linkages: List of Linkages where the transport is located. Defaults to [].
    :vartype linkages: list[Linkage]
    """

    def __post_init__(self):
        _Operation.__post_init__(self)
        self.linkages: list[Linkage] = []

    @property
    def spaces(self) -> list[Location]:
        """Locations at which the process is balanced"""
        return self.linkages

    def writecons_conversion(self, link_times: list[tuple[Linkage, Periods]]):
        """Write the conversion constraints for the transport"""

        def time_checker(res: Resource, loc: Location, time: Periods):
            """This checks if it is actually necessary
            to write conversion at denser temporal scales
            """
            # This checks whether some other aspect is defined at
            # a lower temporal scale

            if loc not in self.model.grb[res]:
                # if not defined for that location, check for a lower order location
                # i.e. location at a lower hierarchy,
                # e.g. say if loc being passed is a city, and a grb has not been defined for it
                # then we need to check at a higher order
                parent = self.space.split(loc)[1]  # get location at one hierarchy above
                if parent:
                    # if that indeed exists, then make the parent the loc
                    # the conversion Balance variables will feature in grb for parent location
                    loc = parent

            _ = self.model.grb[res][loc][time]

            #     self.model.update_grb(resource=res, space=loc, time=time)

            # if time not in self.model.grb[res][loc]:
            #     self.model.update_grb(resource=res, space=loc, time=time)

            if res.inv_of:
                # for inventoried resources, the conversion is written
                # using the time of the base resource's grb
                res = res.inv_of

            times = list(
                [t for t in self.model.grb[res][loc] if self.model.grb[res][loc][t]],
            )
            # write the conversion balance at
            # densest temporal scale in that space
            if times:
                return min(times)

            return time.horizon

        self.conv.balancer()

        if self.conv.pwl:

            conversion = self.conversion[list(self.conversion)[0]]

        else:
            conversion = self.conversion

        shipping_conversion, rest_conversion = {self.conv.base: 1}, {
            k: v for k, v in conversion.items() if k != self.conv.base
        }

        for link_time in link_times:
            link, time = link_time
            source, sink = link.source, link.sink

            # time = link_time[1]

            if link in self.linkages:
                # if the transport is already balanced for the location , Skip
                continue

            for res, par in conversion.items():
                # set, the conversion on the resource
                setattr(res, self.name, self)
                # now there are two cases possible
                # the parameter (par) is positive or negative
                # if positive, the resource is expend
                # if negative, the resource is produced
                # also, the par can be an number or a list of numbers

                # insitu resource (expended and ship_outed within the system)
                # do not initiate a grb so we need to run a check for that first
                if res in self.model.grb:
                    time = time_checker(res, link.source, time)

                    if self.model.grb[res][link][time]:
                        # if the grb has been defined for that resource at that location and time
                        _insitu = False
                    else:
                        _insitu = True
                else:
                    # this implies that the grb needs to be initiated
                    # by declaring relevant variable
                    # the relevant variable will be unbounded
                    _insitu = True

                if isinstance(par, (int | float)) and par < 0:
                    # condition: negative number
                    eff = -par

                    if self.lag:
                        opr = self.operate(link, self.lag.of)
                        rhs = res.expend(self.operate, link.source, self.lag.of)
                    else:
                        opr = self.operate(link, time)
                        rhs = res.expend(opr, link.source, time)

                elif isinstance(par, list) and par[0] < 0:
                    # condition: list with negative numbers
                    eff = [-i for i in par]

                    if self.lag:
                        opr = self.operate(link, self.lag.of)
                        rhs = res.expend(self.operate, link.source, self.lag.of)
                    else:
                        opr = self.operate(link, time)
                        rhs = res.expend(self.operate, link.source, time)

                else:
                    # condition: positive number or list of positive numbers
                    eff = par

                    if self.lag:
                        opr = self.operate(link, self.lag)
                        rhs_export = self.model.ship_out(
                            res,
                            self.operate,
                            link.source,
                            self.lag.of,
                        )
                        rhs_import = self.model.ship_in(
                            res,
                            self.operate,
                            link.sink,
                            self.lag,
                        )
                    else:
                        opr = self.operate(link, time)
                        rhs_export = self.model.ship_out(
                            res,
                            self,
                            link.source,
                            time_checker(res, link.source, time),
                        )
                        rhs_import = self.model.ship_in(
                            res,
                            self,
                            link.sink,
                            time_checker(res, link.sink, time),
                        )

                if _insitu:
                    res.insitu = True
                    _ = rhs_export == True
                    _ = rhs_import == True

                if self.conv.pwl:

                    eff = [conv[res] for conv in self.conversion.values()]

                    if eff[0] < 0:
                        eff = [-i for i in eff]

                    if not self.conv.modes_set:
                        self.model.operate.bound = None
                        _ = opr == dict(enumerate(self.conversion.keys()))

                        self.model.operate.bound = self.conv.model.capacity

                        modes = self.model.modes[-1]
                        self.conv.modes_set = True

                    else:
                        modes = self.conv.modes
                        modes.bind = self.operate
                        self.conv.modes_set = True

                    opr = opr(modes)
                    rhs_export = rhs_export(modes)
                    rhs_import = rhs_import(modes)

                _ = opr[rhs_export] == eff
                _ = opr[rhs_import] == eff

                self.linkages.append(link)

    def __call__(self, resource: Resource | Conversion, lag: Lag = None) -> Conversion:
        """Conversion is called with a Resource to be converted"""
        if not self._conv:

            self.conv = Conversion(operation=self)
            self._conv = True

        self.conv.lag = lag
        return self.conv(resource)



"""Decision Tree"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from itertools import chain
from typing import TYPE_CHECKING, Literal

from .._core._dimension import _Dimension
from ..modeling.indices.domain import Domain
from ..modeling.variables.control import Control
from ..modeling.variables.states import Impact, State, Stream

if TYPE_CHECKING:
    from ..modeling.variables.aspect import Aspect
    from .space import Space
    from .time import Time


@dataclass
class Problem(_Dimension):
    """
    Aspect (Decision) Tree.

    All the aspects are attached to this object.

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, generated based on the class and model name.
    :vartype name: str
    :ivar controls: List of controls. Defaults to empty list.
    :vartype controls: list[Decision]
    :ivar streams: List of streams. Defaults to empty list.
    :vartype streams: list[Stream]
    :ivar impacts: List of consequences. Defaults to empty list.
    :vartype impacts: list[Impact]
    :ivar players: List of players. Defaults to empty list.
    :vartype players: list[Player]

    .. note::
    - name is generated based on Model name
    - controls, streams, impacts, and players are populated as model is defined
    """

    def __post_init__(self):

        _Dimension.__post_init__(self)
        self.states: list[State] = []
        self.controls: list[Control] = []
        self.streams: list[Stream] = []
        self.impacts: list[Impact] = []

    @cached_property
    def time(self) -> Time:
        """Time"""
        return self.model.time

    @cached_property
    def space(self) -> Space:
        """Space"""
        return self.model.space

    @property
    def aspects(self) -> list[Impact | Stream | Control | State]:
        """All Decisions"""
        return self.states + self.controls + self.streams + self.impacts

    @property
    def domains(self) -> list[Domain]:
        """All Domains"""
        return list(set(chain.from_iterable(d.domains for d in self.aspects)))

    def get(
        self,
        keys: Literal[
            "aspects",
            "states",
            "controls",
            "streams",
            "impacts",
            "domains",
        ] = "aspects",
        values: Literal[
            "aspects",
            "states",
            "controls",
            "streams",
            "impacts",
            "domains",
        ] = "domains",
    ) -> dict[Aspect | Domain, list[Aspect | Domain]]:
        """Get a dictionary of the treewith a particular structure

        Args:
            keys (Literal[ 'aspects', 'states', 'controls', 'streams', 'impacts', 'domains' ], optional): Defaults to 'aspects'.
            values (Literal[ 'aspects', 'states', 'controls', 'streams', 'impacts', 'domains' ], optional): Defaults to 'domains'.

        Returns:
            dict[Aspect | Domain, list[Aspect | Domain]]: dictionary with particular structure
        """
        # This function will essentially return a dictionary
        # with any choice of keys and values
        # can come very handy
        if keys in ["aspects", "states", "controls", "streams", "impacts"]:
            keysset: list[Aspect] = getattr(self, keys)

            if values == "domains":
                return {d: [dom.tup for dom in d.domains] for d in keysset if d.domains}
            else:
                return {d: getattr(d, values) for d in keysset if getattr(d, values)}

        if keys in ["domains"]:
            dict_ = {d: [] for d in self.domains}

            for dom in self.domains:
                for val in self.get(keys=values, values=keys):
                    if dom in self.get(keys=values, values=keys)[val]:
                        dict_[dom].append(val)
            return {k.tup: v for k, v in dict_.items() if v}

        raise ValueError(
            "keys must be one of aspects, states, controls, streams, impacts, domains",
        )



"""Inherited _Index class"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from gana import I as Idx
    from gana.block.program import Prg
    from gana.sets.constraint import C

    from ..modeling.indices.domain import Domain
    from ..modeling.variables.aspect import Aspect
    from ..represent.model import Model


@dataclass
class _X(ABC):
    """
    A component (`x`) that functions as an index in the mathematical program.

    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: Citation for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]

    :note:
        - `name` and `model` are set when the component
           is assigned as a Model attribute.
        - `constraints` and `domains` are populated as the program is built.
    """

    label: Optional[str] = None
    citations: Optional[str] = None

    def __post_init__(self):
        # the model
        self.model: Model = None
        # name is given by the model
        self.name: str = ""

        # constraint pnames associated with the component
        self.constraints: list[str] = []
        # domains associated with the component
        self.domains: list[Domain] = []
        # aspects associated with the component with domains
        self.aspects: dict[Aspect, list[Domain]] = {}

    @cached_property
    def program(self) -> Prg:
        """Mathematical program"""
        return self.model.program

    @property
    def cons(self) -> list[C]:
        """Constraints"""
        # this gets the actual constraint objects from the program
        # based on the pname (attribute name) in the program
        return [getattr(self.program, c) for c in self.constraints]

    @property
    @abstractmethod
    def I(self) -> Idx:
        """gana index set"""

    def show(self, descriptive=False, category: Optional[str] = None):
        """Pretty print the component"""
        if category:
            for c in self.cons:
                if c.category and c.category == category:
                    c.show(descriptive)

        else:
            for c in self.cons:
                c.show(descriptive)

    # The reprs are set independently without inheriting _Name
    # which allows a distinction between
    # _Name and _Index when assigned to Model

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __init_subclass__(cls):
        # the hashing will be inherited by the subclasses
        cls.__repr__ = _X.__repr__
        cls.__hash__ = _X.__hash__



"""Consequence"""

from dataclasses import dataclass

from .._core._dimension import _Dimension
from ..components.impact.categories import Economic, Environ, Social


@dataclass
class Consequence(_Dimension):
    """
    A representation of the Consequence of the system dimensions based on the impact
    determined as the product of activity and indicators.

    All impact indicators are attached to this object.

    :param model: Model to which the representation belongs.
    :type model: Model

    :ivar name: Name of the dimension, generated based on the class and model name.
    :vartype name: str
    :ivar envs: List of environmental indicators.
    :vartype envs: list[Env]
    :ivar socs: List of social indicators.
    :vartype socs: list[Soc]
    :ivar ecos: List of economic indicators.
    :vartype ecos: list[Eco]

    .. note::
        - name is self generated
        - environ, socs, and ecos are populated as model is defined

    """

    def __post_init__(self):
        # environmental impact
        self.envs: list[Environ] = []
        # social impact
        self.socs: list[Social] = []
        # economic impact
        self.ecos: list[Economic] = []

        _Dimension.__post_init__(self)

    @property
    def indicators(self):
        """All indicators"""
        return self.envs + self.socs + self.ecos



"""Balance"""

from __future__ import annotations

import time as keep_time
from operator import is_
from typing import TYPE_CHECKING, Self

from ...components.commodity.stored import Stored

if TYPE_CHECKING:
    from gana.sets.constraint import C

    from ..._core._x import _X
    from ...components.spatial.linkage import Linkage
    from ...components.spatial.location import Location
    from ...components.temporal.periods import Periods
    from ..indices.domain import Domain
    from ..variables.aspect import Aspect


class Balance:
    """Performs a general commodity balance

    Args:
        aspect (Aspect. optional): Aspect to which the constraint is applied
        domain (Domain. optional): Domain over which the aspect is defined

    Attributes:
        name (str. optional): Name of the constraint.
    """

    def __init__(self, aspect: Aspect, domain: Domain, label: str = ""):
        self.aspect = aspect
        self.domain = domain
        self.label = label

        self.model = self.aspect.model
        self.program = self.model.program
        self.grb = self.model.grb

        if self.domain.modes:
            return

        commodity = self.domain.commodity
        binds = self.domain.binds
        time = self.domain.time

        loc = (
            self.domain.linkage.source
            if self.aspect.sign == -1 and self.domain.linkage
            else (
                self.domain.linkage.sink
                if self.domain.linkage
                else self.domain.location
            )
        )

        _ = self.grb[commodity][loc][time]

        if not binds and commodity:
            # if no binds, then create GRB or append to exisiting GRB
            # writecons_grb will figure it out
            self.writecons_grb(commodity, loc, time)
            return

        # if there are binds
        if commodity.insitu:
            # # we need to still check if this is this is an insitu (e.g. a storage commodity)
            # if the commodity is insitu that means that
            # no external bounds have been defined
            # a GRB is still needed

            self.writecons_grb(commodity, loc, time)

        if self.aspect(commodity, loc, time) not in self.grb[commodity][loc][time]:
            # for the second check, consider the case where

            # # these do not get their own GRB, as they are only utilized within a process

            # if there is already a GRB existing
            # add the bind to the GRB at the same scale
            self.writecons_grb(commodity, loc, time)

    @property
    def name(self) -> str:
        """Name of the constraint"""
        return f"{self.aspect.name}{self.domain}"

    @property
    def mapped_to(self) -> list[Domain]:
        """List of domains that the aspect has been mapped to"""
        return self.aspect.maps

    @property
    def sign(self) -> float:
        """Returns the aspect"""
        return self.aspect.sign

    def _update_constraint(
        self,
        name: str,
        stored: bool,
        time: Periods,
        cons_grb: C,
    ) -> bool:
        """Updates an existing GRB constraint with the new aspect"""
        if stored and self.aspect.name == "inventory":

            if len(time) == 1:
                return False

            # if inventory is being add to GRB
            lagged_domain = self.domain.change({"lag": -1 * time, "periods": None})

            setattr(
                self.program,
                name,
                cons_grb + self(*lagged_domain).V() - self(*self.domain).V(),
            )
        else:
            setattr(
                self.program,
                name,
                (
                    cons_grb + self(*self.domain).V()
                    if self.aspect.ispos
                    else cons_grb - self(*self.domain).V()
                ),
            )

        return True

    def _create_constraint(
        self, name: str, stored: bool, time: Periods, space: Location | Linkage
    ) -> bool:
        """Creates a new GRB constraint"""

        if stored and self.aspect.name == "inventory":

            if len(time) == 1:
                return False

            # TODO: potential bug here
            # this avoids writing inv_ntw_t - inv_ntw_t-1 = 0
            # the need for this check is absurd
            # it ensures that inventory balances are not created for the network directly
            # when location-wise inventory balances can be written
            # this is a temporary fix and needs to be generalized
            if self.model.network.has and space.isnetwork:
                # if the location is a network, we cannot have storage
                return False

            # if inventory is being add to GRB
            lagged_domain = self.domain.change({"lag": -1 * time, "periods": None})
            cons_grb = -self(*self.domain).V() + self(*lagged_domain).V() == 0

        else:
            cons_grb = (
                self(*self.domain).V() == 0
                if self.aspect.ispos
                else -self(*self.domain).V() == 0
            )

        cons_grb.categorize("General Resource Balance")

        setattr(
            self.program,
            name,
            cons_grb,
        )

        return True

    def writecons_grb(self, commodity, loc, time):
        """Writes the stream balance constraint"""

        if (
            loc.isin is not None
            and not self.grb[commodity][loc][time]
            and self.grb[commodity][loc.isin][time]
        ):
            # if the location is in another location
            # and there is no GRB for that location
            # work with the parent location
            loc = loc.isin

        _name = f"{commodity}_{loc}_{time}_grb"

        if isinstance(commodity, Stored):
            stored = True

        else:
            stored = False

            if self.grb[commodity][loc]:
                # If a GRB exists at a lower temporal order, append to that
                lower_times = [t for t in self.grb[commodity][loc] if t > time]
                if lower_times:
                    _ = self.aspect(commodity, loc, lower_times[0]) == True
                    return

        # ---- initialize GRB for commodity if necessary -----

        if not self.grb[commodity][loc][time]:
            # this checks whether a general commodity balance has been defined
            # for the commodity in that space and time

            print(
                f"--- General Resource Balance for {commodity} in ({loc}, {time}): initializing constraint, adding {self.aspect}{self.domain}",
            )

            start = keep_time.time()

            made = self._create_constraint(_name, stored, time, loc)

        # ---- add aspect to GRB if not added already ----

        # elif self not in self.grb[commodity][loc][time]:

        else:

            print(
                f"--- General Resource Balance for {commodity} in ({loc}, {time}): adding {self.aspect}{self.domain}",
            )

            start = keep_time.time()

            made = self._update_constraint(
                _name, stored, time, getattr(self.program, _name)
            )

        if not made:
            return

        end = keep_time.time()
        print(f"    Completed in {end-start} seconds")

        # updates the constraints in all the indices of self.domain
        # add constraint name to aspect
        self.domain.update_cons(_name)

        if _name not in self.aspect.constraints:
            self.aspect.constraints.append(_name)

        # update the GRB aspects
        self.grb[commodity][loc][time].append(self)

    def __eq__(self, other: Self):
        return is_(self.aspect, other.aspect) and self.domain == other.domain

    def __call__(self, *index: _X):
        """Returns the variable for the aspect at the given index"""
        return self.aspect(*index)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)



"""Player"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..._core._component import _Component

if TYPE_CHECKING:
    pass


@dataclass
class Player(_Component):
    """Player or Actor, the one taking the decisions
    based on information provided

    Players own certain processes and be responsible for the streams and impact
    caused by their decisions pertaining to this

    :param basis: Unit basis of the component. Defaults to None.
    :type basis: Unit, optional
    :param label: An optional label for the component. Defaults to None.
    :type label: str, optional
    :param captions: An optional citation or description for the component. Defaults to None.
    :type captions: str | list[str] | dict[str, str | list[str]], optional

    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str

    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    """

    def __post_init__(self):
        _Component.__post_init__(self)



"""Decorators for functions"""


def once(func):
    """Ensures the function is executed only once"""

    def wrapper(*args, **kwargs):
        if not hasattr(wrapper, "has_run"):
            wrapper.has_run = True
            return func(*args, **kwargs)

    return wrapper



"""Model classifiers"""

from enum import Enum


class Uncertainty(Enum):
    """Model uncertainty"""

    DETERMINISTIC = "Deterministic (certain inputs)"
    STOCHASTIC = "Stochastic (random inputs)"
    PARAMETRIC = "Parametric (uncertain parameters)"


class Structure(Enum):
    """Model Mathematical Structure/Order/Variable Types"""

    LINEAR = "Linear"
    NONLINEAR = "Nonlinear"
    INTEGER = "Integer"
    MIXED_INTEGER = "Mixed-Integer"


class Scale(Enum):
    """Model Scale/Order"""

    MULTI = "Multi"
    SINGLE = "Single"


class Paradigm(Enum):
    """Modeling Paradigm"""

    DESIGN = "Design"
    SCHEDULING = "Scheduling"
    CONTROL = "Control"



"""Inherited _Name (reprs) class"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Optional


@dataclass
class _Name:
    """
    Inherited by components that only have a name.

    These do not feature as indices in the model
    nor do they have modeling aspects.

    For components with a:
        - mathematical program, use `X`
        - mathematical program and
          input parameters (modeling aspects), use `Component`

    :param label: Label used for plotting. Defaults to None.
    :type label: str, optional

    :ivar name: Name of the object. Defaults to ''.
    :vartype name: str

    :note:
        - `name` is set when the component is made a Model attribute.
    """

    label: Optional[str] = None

    def __post_init__(self):
        self.name = ""

    # -----------------------------------------------------
    #                    Hashing
    # -----------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def __init_subclass__(cls):
        cls.__repr__ = _Name.__repr__
        cls.__hash__ = _Name.__hash__



"""Commodity Base Class"""

from dataclasses import dataclass
from typing import Self

from ..._core._component import _Component
from ...modeling.parameters.conversion import Conversion


@dataclass
class _Commodity(_Component):
    """
    A commodity, can be a material, chemical, energy, etc.

    :param label: Label of the commodity, used for plotting. Defaults to None.
    :type label: str, optional
    :param name: Name of the commodity, used for indexing. Defaults to None.
    :type name: str, optional
    :param basis: Unit basis of the commodity. Defaults to None.
    :type basis: Unit, optional


    :ivar model: The model to which the component belongs.
    :vartype model: Model
    :ivar name: Set when the component is assigned as a Model attribute.
    :vartype name: str
    :ivar constraints: List of constraints associated with the component.
    :vartype constraints: list[str]
    :ivar domains: List of domains associated with the component.
    :vartype domains: list[Domain]
    :ivar aspects: Aspects associated with the component with domains.
    :vartype aspects: dict[Aspect, list[Domain]]
    :ivar conversions: List of conversions associated with the commodity. Defaults to [].
    :vartype conversions: list[Conversion]
    :ivar insitu: If the commodity only exists insitu, i.e., does not scale any domains
    :vartype insitu: bool, optional
    """

    def __post_init__(self):
        _Component.__post_init__(self)

        # list of conversions associated with the commodity
        self.conversions: list[Conversion] = []

        # flag indicates whether the commodity is produced and expended insitu only
        self.insitu = False

    @property
    def conversion(self) -> dict[Conversion, int | float]:
        """Conversion"""
        # if no conversion is set, return 1.0
        if not self.conversions:
            return {self: 1.0}
        # if there is a conversion, return its parameter in all tasks its
        # associated with
        return {task: task.conversion[self] for task in self.conversions}

    def __setattr__(self, name, value):

        if isinstance(value, Conversion):
            # update conversions
            self.conversions.append(value)

        super().__setattr__(name, value)

    def __mul__(self, other: int | float) -> Conversion:
        # multiplying a number with a resources gives conversion
        # math operations with conversions form the balance in tasks
        conv = Conversion()
        conv.conversion = {self: other}
        return conv

    def __rmul__(self, other: int | float) -> Conversion:
        # reverse multiplication
        return self * other

    def __add__(self, other: Conversion) -> Conversion:
        conv = Conversion()
        if isinstance(other, _Commodity):
            # if another resource is added, give it the parameter 1
            conv.conversion = {self: 1, other: 1}
            return conv

        # if added with another conversion, updated the balance
        conv.conversion = {self: 1, **other.conversion}
        return conv

    def __neg__(self) -> Conversion:
        # just multiply by -1
        return self * -1

    def __sub__(self, other: Conversion | Self):
        if isinstance(other, _Commodity):
            # if another resource is subtracted
            # give it the parameter -1
            return self + -1 * other
        if isinstance(other, Conversion):
            # if another conversion is subtracted, update the balance
            conv = Conversion()

            conv.conversion = {
                self: 1,
                **{
                    res: (
                        -1 * par if isinstance(par, (int, float)) else [-i for i in par]
                    )
                    for res, par in other.conversion.items()
                },
            }
            return conv

    def __truediv__(self, other: int | float):
        # treat division as multiplication by the inverse
        return self * (1 / other)
